import logging
from typing import Union

import typer

from runhouse.rh_config import configs, rns_client

logger = logging.getLogger(__name__)


def is_interactive():
    import __main__ as main

    return not hasattr(main, "__file__")


def login(
    token: str = None,
    download_config: bool = None,
    upload_config: bool = None,
    download_secrets: bool = None,
    upload_secrets: bool = None,
    ret_token: bool = False,
    interactive: bool = False,
):
    """Login to Runhouse. Validates token provided, with options to upload or download stored secrets or config between
    local environment and Runhouse / Vault.
    """
    from runhouse import Secrets

    if is_interactive() or interactive:
        from getpass import getpass

        from rich.console import Console

        console = Console()
        console.print(
            """
            ____              __                             @ @ @
           / __ \__  ______  / /_  ____  __  __________     []___
          / /_/ / / / / __ \/ __ \/ __ \/ / / / ___/ _ \   /    /\____    @@
         / _, _/ /_/ / / / / / / / /_/ / /_/ (__  )  __/  /_/\_//____/\  @@@@
        /_/ |_|\__,_/_/ /_/_/ /_/\____/\__,_/____/\___/   | || |||__|||   ||
        """
        )
        link = (
            f'[link={configs.get("api_server_url")}/dashboard/?option=token]https://api.run.house[/link]'
            if is_interactive()
            else f'{configs.get("api_server_url")}/dashboard/?option=token'
        )
        console.print(
            f"Retrieve your token :key: here to use :person_running: :house: Runhouse for "
            f"secrets and artifact management: {link}",
            style="bold yellow",
        )
        if not token:
            token = getpass("Token: ")

        download_config = (
            download_config
            if download_config is not None
            else typer.confirm(
                "Download config from Runhouse to your local .rh folder?"
            )
        )
        download_secrets = (
            download_secrets
            if download_secrets is not None
            else typer.confirm(
                "Download secrets from Vault to your local Runhouse config?"
            )
        )
        upload_config = (
            upload_config
            if upload_config is not None
            else typer.confirm("Upload your local config to Runhouse?")
        )
        upload_secrets = (
            upload_secrets
            if upload_secrets is not None
            else typer.confirm("Upload your enabled cloud provider secrets to Vault?")
        )

    if token:
        configs.set("token", token)

    if download_config:
        configs.download_and_save_defaults()
        # We need to fresh the RNSClient to use the newly loaded configs
        rns_client.refresh_defaults()
    elif upload_config:
        configs.upload_defaults(defaults=configs.defaults_cache)
    else:
        # If we are not downloading or uploading config, we still want to make sure the token is valid
        try:
            configs.download_defaults()
        except:
            logger.error("Failed to validate token")
            return None

    if download_secrets:
        Secrets.download_into_env()

    if upload_secrets:
        Secrets.extract_and_upload(interactive=interactive)

    logger.info("Successfully logged into Runhouse.")
    if ret_token:
        return token


def logout(
    delete_loaded_secrets: bool = None,
    delete_rh_config_file: bool = None,
    interactive: bool = None,
):
    """Logout from Runhouse. Provides option to delete credentials from the Runhouse config and the underlying
     credentials file. Token is also deleted from the config.

    Args:
        delete_loaded_secrets (bool, optional): If True, deletes the provider credentials file. Defaults to None.
        delete_rh_config_file (bool, optional): If True, deletes the rh config file. Defaults to None.
        interactive (bool, optional): If True, runs the logout process in interactive mode. Defaults to None.

    Returns:
        None
    """
    from runhouse import Secrets

    interactive_session: bool = (
        interactive if interactive is not None else is_interactive()
    )
    for provider in Secrets.enabled_providers():
        provider_name: str = provider.PROVIDER_NAME
        provider_creds_path: Union[str, tuple] = provider.default_credentials_path()

        if interactive_session:
            delete_loaded_secrets = typer.confirm(
                f"Delete credentials file for {provider_name}?"
            )

        configs.delete(provider_name)

        if delete_loaded_secrets:
            provider.delete_secrets_file(provider_creds_path)
            logger.info(
                f"Deleted {provider_name} credentials file from path: {provider_creds_path}"
            )

    # Delete token from rh config file
    configs.delete(key="token")

    rh_config_path = configs.CONFIG_PATH
    if not delete_rh_config_file and interactive_session:
        delete_rh_config_file = typer.confirm("Delete your local Runhouse config file?")

    if delete_rh_config_file:
        # Delete the credentials file on the file system
        configs.delete_defaults(rh_config_path)
        logger.info(f"Deleted Runhouse config file from path: {rh_config_path}")

    logger.info("Successfully logged out of Runhouse.")



import ast
import datetime
import json
import uuid

from requests import Response


def timing(func):
    def wrapper(*args, **kwargs):
        import time

        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(f"Finished {func.__name__.title()} in {int((end - start))} seconds")
        return result

    return wrapper


def remove_null_values_from_dict(source_dic: dict) -> dict:
    return {k: v for k, v in source_dic.items() if v is not None}


def load_resp_content(resp: Response) -> dict:
    return json.loads(resp.content)


def read_resp_data(resp: Response):
    return load_resp_content(resp).get("data", {})


def to_bool(value):
    try:
        return ast.literal_eval(value)
    except:
        return value


def is_jsonable(myjson):
    try:
        json.dumps(myjson)
    except ValueError:
        return False
    return True


def generate_uuid():
    return uuid.uuid4().hex


def log_timestamp():
    return int(datetime.datetime.now(datetime.timezone.utc).timestamp() * 1e9)


def create_s3_bucket(bucket_name: str):
    """Create bucket in S3 if it does not already exist."""
    from sky.data.storage import S3Store

    s3_store = S3Store(name=bucket_name, source="")
    return s3_store


def create_gcs_bucket(bucket_name: str):
    """Create bucket in GS if it does not already exist."""
    from sky.data.storage import GcsStore

    gcs_store = GcsStore(name=bucket_name, source="")
    return gcs_store



import copy
import logging
import os
from pathlib import Path
from typing import Dict, List, Optional

import fsspec
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import ray.data

import runhouse as rh
from runhouse.rh_config import rns_client
from runhouse.rns.folders.folder import folder
from runhouse.rns.obj_store import _current_cluster

from .. import OnDemandCluster, Resource

logger = logging.getLogger(__name__)


class Table(Resource):
    RESOURCE_TYPE = "table"
    DEFAULT_FOLDER_PATH = "/runhouse-table"
    DEFAULT_CACHE_FOLDER = ".cache/runhouse/tables"
    DEFAULT_STREAM_FORMAT = "pyarrow"
    DEFAULT_BATCH_SIZE = 256
    DEFAULT_PREFETCH_BLOCKS = 1

    def __init__(
        self,
        path: str,
        name: Optional[str] = None,
        file_name: Optional[str] = None,
        system: Optional[str] = None,
        data_config: Optional[dict] = None,
        dryrun: bool = False,
        partition_cols: Optional[List] = None,
        stream_format: Optional[str] = None,
        metadata: Optional[Dict] = None,
        **kwargs,
    ):
        """
        The Runhouse Table object.

        .. note::
            To build a Table, please use the factory method :func:`table`.
        """
        super().__init__(name=name, dryrun=dryrun)
        self.file_name = file_name

        # Use factory method so correct subclass for system is returned
        # strip filename from path if provided
        self._folder = folder(
            path=str(Path(path).parents[0]) if Path(path).suffix else path,
            system=system,
            data_config=data_config,
            dryrun=dryrun,
        )

        self._cached_data = None
        self.partition_cols = partition_cols
        self.stream_format = stream_format or self.DEFAULT_STREAM_FORMAT
        self.metadata = metadata or {}

    @staticmethod
    def from_config(config: dict, dryrun=True):
        if isinstance(config["system"], dict):
            config["system"] = OnDemandCluster.from_config(
                config["system"], dryrun=dryrun
            )
        return Table(**config, dryrun=dryrun)

    @property
    def config_for_rns(self):
        config = super().config_for_rns
        if isinstance(self._folder, Resource):
            config["system"] = self._resource_string_for_subconfig(self.system)
            config["data_config"] = self._folder._data_config
        else:
            config["system"] = self.system
        self.save_attrs_to_config(
            config, ["path", "partition_cols", "metadata", "file_name"]
        )
        config.update(config)

        return config

    @property
    def data(self) -> ray.data.Dataset:
        """Get the table data. If data is not already cached, return a Ray dataset.

        With the dataset object we can stream or convert to other types, for example:

        .. code-block:: python

            data.iter_batches()
            data.to_pandas()
            data.to_dask()
        """
        if self._cached_data is not None:
            return self._cached_data
        return self.read_ray_dataset()

    @data.setter
    def data(self, new_data):
        """Update the data blob to new data"""
        self._cached_data = new_data
        # TODO should we save here?
        # self.save(overwrite=True)

    @property
    def system(self):
        return self._folder.system

    @system.setter
    def system(self, new_system):
        self._folder.system = new_system

    @property
    def path(self):
        if self.file_name:
            return f"{self._folder.path}/{self.file_name}"
        return self._folder.path

    @path.setter
    def path(self, new_path):
        self._folder.path = new_path

    def set_metadata(self, key, val):
        self.metadata[key] = val

    def get_metadata(self, key):
        return self.metadata.get(key)

    @property
    def fsspec_url(self):
        if self.file_name:
            return f"{self._folder.fsspec_url}/{self.file_name}"
        return self._folder.fsspec_url

    @property
    def data_config(self):
        return self._folder.data_config

    @data_config.setter
    def data_config(self, new_data_config):
        self._folder.data_config = new_data_config

    @classmethod
    def from_name(cls, name, dryrun=True):
        """Load existing Table via its name."""
        config = rns_client.load_config(name=name)
        if not config:
            raise ValueError(f"Table {name} not found.")
        # We don't need to load the cluster dict here (if system is a cluster) because the table init
        # goes through the Folder factory method, which does that.

        # Uses the table subclass associated with the `resource_subtype`
        table_cls = _load_table_subclass(config=config, dryrun=dryrun)
        return table_cls.from_config(config=config, dryrun=dryrun)

    def to(self, system, path=None, data_config=None):
        """Copy and return the table on the given filesystem and path."""
        new_table = copy.copy(self)
        new_table._folder = self._folder.to(
            system=system, path=path, data_config=data_config
        )
        return new_table

    def _save_sub_resources(self):
        if isinstance(self.system, Resource):
            self.system.save()

    def write(self):
        """Write underlying table data to fsspec URL."""
        if self._cached_data is not None:
            data_to_write = self.data

            if isinstance(data_to_write, pd.DataFrame):
                data_to_write = self.ray_dataset_from_pandas(data_to_write)

            if isinstance(data_to_write, pa.Table):
                data_to_write = self.ray_dataset_from_arrow(data_to_write)

            if not isinstance(data_to_write, ray.data.Dataset):
                raise TypeError(f"Invalid Table format {type(data_to_write)}")

            self.write_ray_dataset(data_to_write)
            logger.info(f"Saved {str(self)} to: {self.fsspec_url}")

        return self

    def fetch(self, columns: Optional[list] = None) -> pa.Table:
        """Returns the complete table contents."""
        # https://arrow.apache.org/docs/python/generated/pyarrow.parquet.read_table.html
        self._cached_data = self.read_table_from_file(columns)
        if self._cached_data is not None:
            return self._cached_data

        # When trying to read a file like object could lead to IsADirectoryError if the folder path is actually a
        # directory and the file has been automatically generated for us inside the folder
        # (ex: with pyarrow table or with partitioned data that saves multiple files within the directory)

        try:
            table_data = pq.read_table(
                self.path, columns=columns, filesystem=self._folder.fsspec_fs
            )

            if not table_data:
                raise ValueError(f"No table data found in path: {self.path}")

            self._cached_data = table_data
            return self._cached_data

        except:
            raise Exception(f"Failed to read table in path: {self.path}")

    def __getstate__(self):
        """Override the pickle method to clear _cached_data before pickling."""
        state = self.__dict__.copy()
        state["_cached_data"] = None
        return state

    def __iter__(self):
        for block in self.stream(batch_size=self.DEFAULT_BATCH_SIZE):
            for sample in block:
                yield sample

    def __len__(self):
        if isinstance(self.data, pd.DataFrame):
            len_dataset = self.data.shape[0]

        elif isinstance(self.data, ray.data.Dataset):
            len_dataset = self.data.count()

        else:
            if not hasattr(self.data, "__len__") or not self.data:
                raise RuntimeError("Cannot get len for dataset.")
            else:
                len_dataset = len(self.data)

        return len_dataset

    def __str__(self):
        return self.__class__.__name__

    def stream(
        self,
        batch_size: int,
        drop_last: bool = False,
        shuffle_seed: Optional[int] = None,
        shuffle_buffer_size: Optional[int] = None,
        prefetch_blocks: Optional[int] = None,
    ):
        """Return a local batched iterator over the ray dataset."""
        ray_data = self.data

        if self.stream_format == "torch":
            # https://docs.ray.io/en/master/data/api/doc/ray.data.Dataset.iter_torch_batches.html#ray.data.Dataset.iter_torch_batches
            return ray_data.iter_torch_batches(
                batch_size=batch_size,
                prefetch_blocks=prefetch_blocks or self.DEFAULT_PREFETCH_BLOCKS,
                drop_last=drop_last,
                local_shuffle_buffer_size=shuffle_buffer_size,
                local_shuffle_seed=shuffle_seed,
            )

        elif self.stream_format == "tf":
            # https://docs.ray.io/en/master/data/api/doc/ray.data.Dataset.iter_tf_batches.html
            return ray_data.iter_tf_batches(
                batch_size=batch_size,
                prefetch_blocks=prefetch_blocks or self.DEFAULT_PREFETCH_BLOCKS,
                drop_last=drop_last,
                local_shuffle_buffer_size=shuffle_buffer_size,
                local_shuffle_seed=shuffle_seed,
            )
        else:
            # https://docs.ray.io/en/latest/data/api/dataset.html#ray.data.Dataset.iter_batches
            return ray_data.iter_batches(
                batch_size=batch_size,
                prefetch_blocks=prefetch_blocks or self.DEFAULT_PREFETCH_BLOCKS,
                batch_format=self.stream_format,
                drop_last=drop_last,
                local_shuffle_buffer_size=shuffle_buffer_size,
                local_shuffle_seed=shuffle_seed,
            )

    def read_ray_dataset(self, columns: Optional[List[str]] = None):
        """Read parquet data as a ray dataset object"""
        # https://docs.ray.io/en/latest/data/api/input_output.html#parquet
        dataset = ray.data.read_parquet(
            self.fsspec_url, columns=columns, filesystem=self._folder.fsspec_fs
        )
        return dataset

    def write_ray_dataset(self, data_to_write: ray.data.Dataset):
        """Write a ray dataset to a fsspec filesystem"""
        if self.partition_cols:
            # TODO [JL]: https://arrow.apache.org/docs/python/generated/pyarrow.parquet.write_to_dataset.html
            logger.warning("Partitioning by column not currently supported.")
            pass

        # https://docs.ray.io/en/master/data/api/doc/ray.data.Dataset.write_parquet.html
        data_to_write.repartition(os.cpu_count() * 2).write_parquet(
            self.fsspec_url, filesystem=self._folder.fsspec_fs
        )

    @staticmethod
    def ray_dataset_from_arrow(data: pa.Table):
        """Convert an Arrow Table to a Ray Dataset"""
        return ray.data.from_arrow(data)

    @staticmethod
    def ray_dataset_from_pandas(data: pd.DataFrame):
        """Convert an Pandas DataFrame to a Ray Dataset"""
        return ray.data.from_pandas(data)

    def read_table_from_file(self, columns: Optional[list] = None):
        try:
            with fsspec.open(self.fsspec_url, mode="rb", **self.data_config) as t:
                table_data = pq.read_table(t.full_name, columns=columns)
            return table_data
        except:
            return None

    def delete_in_system(self, recursive: bool = True):
        """Remove contents of all subdirectories (ex: partitioned data folders)"""
        # If file(s) are directories, recursively delete contents and then also remove the directory
        # https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.spec.AbstractFileSystem.rm
        self._folder.fsspec_fs.rm(self.path, recursive=recursive)
        if self.system == "file" and recursive:
            self._folder.delete_in_system()

    def exists_in_system(self):
        """Whether table exists in file system"""
        return (
            self._folder.exists_in_system()
            and len(self._folder.ls(self.fsspec_url)) >= 1
        )


def _load_table_subclass(config: dict, dryrun: bool, data=None):
    """Load the relevant Table subclass based on the config or data type provided"""
    resource_subtype = config.get("resource_subtype", Table.__name__)

    try:
        import datasets

        if resource_subtype == "HuggingFaceTable" or isinstance(data, datasets.Dataset):
            from .huggingface_table import HuggingFaceTable

            return HuggingFaceTable.from_config(config, dryrun=dryrun)
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    try:
        if resource_subtype == "PandasTable" or isinstance(data, pd.DataFrame):
            from .pandas_table import PandasTable

            return PandasTable.from_config(config, dryrun=dryrun)
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    try:
        import dask.dataframe as dd

        if resource_subtype == "DaskTable" or isinstance(data, dd.DataFrame):
            from .dask_table import DaskTable

            return DaskTable.from_config(config, dryrun=dryrun)
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    try:
        import ray

        if resource_subtype == "RayTable" or isinstance(data, ray.data.Dataset):
            from .ray_table import RayTable

            return RayTable.from_config(config, dryrun=dryrun)
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    try:
        import cudf

        if resource_subtype == "CudfTable" or isinstance(data, cudf.DataFrame):
            raise NotImplementedError("Cudf not currently supported")
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    if resource_subtype == "Table" or isinstance(data, pa.Table):
        new_table = Table.from_config(config, dryrun)
        return new_table
    else:
        raise TypeError(
            f"Unsupported data type {type(data)} for Table construction. "
            f"For converting data to pyarrow see: "
            f"https://arrow.apache.org/docs/7.0/python/generated/pyarrow.Table.html"
        )


def table(
    data=None,
    name: Optional[str] = None,
    path: Optional[str] = None,
    system: Optional[str] = None,
    data_config: Optional[dict] = None,
    partition_cols: Optional[list] = None,
    mkdir: bool = False,
    dryrun: bool = False,
    stream_format: Optional[str] = None,
    metadata: Optional[dict] = None,
    load: bool = True,
) -> Table:
    """Constructs a Table object, which can be used to interact with the table at the given path.

    Args:
        data: Data to be stored in the table.
        name (Optional[str]): Name for the table, to reuse it later on.
        path (Optional[str]): Full path to the data file.
        system (Optional[str]): File system. Currently this must be one of:
            [``file``, ``github``, ``sftp``, ``ssh``,``s3``, ``gs``, ``azure``].
        data_config (Optional[dict]): The data config to pass to the underlying fsspec handler.
        partition_cols (Optional[list]): List of columns to partition the table by.
        mkdir (bool): Whether to (Default: ``False``)
        dryrun (bool): Whether to create the Table if it doesn't exist, or load a Table object as a dryrun.
            (Default: ``False``)
        stream_format (Optional[str]): Format to stream the Table as.
            Currently this must be one of: [``pyarrow``, ``torch``, ``tf``, ``pandas``]
        metadata (Optional[dict]): Metadata to store for the table.
        load (bool): Whether to load an existing config for the Table. (Default: ``True``)

    Returns:
        Table: The resulting Table object.

    Example:
        >>> # Create and save (pandas) table
        >>> rh.table(
        >>>    data=data,
        >>>    name="~/my_test_pandas_table",
        >>>    path="table_tests/test_pandas_table.parquet",
        >>>    system="file",
        >>>    mkdir=True,
        >>> )
        >>>
        >>> # Load table from above
        >>> reloaded_table = rh.table(name="~/my_test_pandas_table")
    """
    config = rns_client.load_config(name) if load else {}

    config["system"] = (
        system
        or config.get("system")
        or _current_cluster(key="config")
        or rns_client.DEFAULT_FS
    )

    name = name or config.get("rns_address") or config.get("name")

    data_path = path or config.get("path")
    file_name = None
    if data_path:
        # Extract the file name from the path if provided
        full_path = Path(data_path)
        file_suffix = full_path.suffix
        if file_suffix:
            data_path = str(full_path.parent)
            file_name = full_path.name

    if data_path is None:
        # If no path is provided we need to create one based on the name of the table
        table_name_in_path = rns_client.resolve_rns_data_resource_name(name)
        if (
            config["system"] == rns_client.DEFAULT_FS
            or config["system"] == _current_cluster()
            or (
                isinstance(config["system"], dict)
                and config["system"]["name"] == _current_cluster()
            )
        ):
            # create random path to store in .cache folder of local filesystem
            data_path = str(
                Path(
                    f"~/{Table.DEFAULT_CACHE_FOLDER}/{table_name_in_path}"
                ).expanduser()
            )
        else:
            # save to the default bucket
            data_path = f"{Table.DEFAULT_FOLDER_PATH}/{table_name_in_path}"

    if isinstance(config["system"], str) and rns_client.exists(
        config["system"], resource_type="cluster"
    ):
        config["system"] = rns_client.load_config(config["system"])
    elif isinstance(config["system"], dict):
        from runhouse.rns.hardware.cluster import Cluster

        config["system"] = Cluster.from_config(config["system"])

    config["name"] = name
    config["path"] = data_path
    config["file_name"] = file_name or config.get("file_name")
    config["data_config"] = data_config or config.get("data_config")
    config["partition_cols"] = partition_cols or config.get("partition_cols")
    config["stream_format"] = stream_format or config.get("stream_format")
    config["metadata"] = metadata or config.get("metadata")

    if mkdir:
        # create the remote folder for the table
        rh.folder(path=data_path, system=config["system"], dryrun=True).mkdir()

    new_table = _load_table_subclass(config=config, dryrun=dryrun, data=data)
    if data is not None:
        new_table.data = data

    return new_table



import logging
import os
from pathlib import Path
from typing import Optional

from runhouse import Secrets

logger = logging.getLogger(__name__)


class SSHSecrets(Secrets):
    PROVIDER_NAME = "ssh"
    CREDENTIALS_FILE = os.path.expanduser("~/.ssh")

    @classmethod
    def read_secrets(cls, from_env: bool = False, file_path: Optional[str] = None):
        if from_env:
            raise NotImplementedError(
                "SSH secrets cannot be read from environment variables."
            )
        else:
            creds_path = Path(file_path or cls.CREDENTIALS_FILE).expanduser()
            config_data = {}
            for f in creds_path.glob("*"):
                # TODO do we need to support pem files?
                if f.suffix != ".pub":
                    continue
                if f.name == "sky-key.pub":
                    # We don't need to store duplicate ssh keys for sky (Sky is already a builtin provider)
                    continue
                if not Path(creds_path / f.stem).exists():
                    logger.warning(
                        f"Private key {f.stem} not found for public key {f.name}, skipping."
                    )
                    continue
                # Grab public key
                config_data[f.name] = Path(creds_path / f).read_text()
                # Grab corresponding private key
                config_data[f.stem] = Path(creds_path / f.stem).read_text()

        return config_data

    @classmethod
    def save_secrets(cls, secrets: dict, overwrite: bool = False):
        dest_path = Path(cls.default_credentials_path()).expanduser()
        cls.check_secrets_for_mismatches(
            secrets_to_save=secrets, secrets_path=str(dest_path), overwrite=overwrite
        )

        dest_path.mkdir(parents=True, exist_ok=True)

        valid_secrets = {}
        for key_name, key in secrets.items():
            if key_name == "provider":
                continue
            key_path = dest_path / key_name
            if key_path.exists():
                logger.warning(f"Key {key_name} already exists, skipping.")
                continue
            key_path.write_text(key)
            key_path.chmod(0o600)
            valid_secrets[key_name] = str(key_path)

        cls.add_provider_to_rh_config(secrets_for_config={cls.PROVIDER_NAME: secrets})



import os
from pathlib import Path
from typing import Optional

from runhouse import Secrets


class LambdaSecrets(Secrets):
    PROVIDER_NAME = "lambda"
    CREDENTIALS_FILE = os.path.expanduser("~/.lambda_cloud/lambda_keys")

    @classmethod
    def read_secrets(cls, from_env: bool = False, file_path: Optional[str] = None):
        if from_env:
            raise NotImplementedError(
                "Lambda secrets cannot be read from environment variables."
            )
        else:
            from sky.clouds.lambda_cloud import lambda_utils

            client = lambda_utils.LambdaCloudClient()
            api_key = client.api_key

        return {"api_key": api_key}

    @classmethod
    def save_secrets(cls, secrets: dict, overwrite: bool = False):

        dest_path = cls.default_credentials_path()
        cls.check_secrets_for_mismatches(
            secrets_to_save=secrets, secrets_path=dest_path, overwrite=overwrite
        )

        Path(dest_path).parent.mkdir(parents=True, exist_ok=True)

        with open(dest_path, "w") as f:
            f.write(f'api_key = {secrets["api_key"]}\n')

        cls.add_provider_to_rh_config()



import json
import logging
import traceback
from concurrent import futures
from pathlib import Path

import grpc
import ray
import ray.cloudpickle as pickle
import requests
from sky.skylet.autostop_lib import set_last_active_time_to_now

import runhouse.servers.grpc.unary_pb2 as pb2

import runhouse.servers.grpc.unary_pb2_grpc as pb2_grpc
from runhouse.rh_config import configs, obj_store
from runhouse.rns.packages.package import Package
from runhouse.rns.run_module_utils import call_fn_by_type, get_fn_by_name
from runhouse.rns.top_level_rns_fns import (
    clear_pinned_memory,
    pinned_keys,
    remove_pinned_object,
)
from runhouse.servers.grpc.unary_client import OutputType

logger = logging.getLogger(__name__)


class UnaryService(pb2_grpc.UnaryServicer):
    DEFAULT_PORT = 50052
    MAX_MESSAGE_LENGTH = 1 * 1024 * 1024 * 1024  # 1 GB
    LOGGING_WAIT_TIME = 1.0
    SKY_YAML = str(Path("~/.sky/sky_ray.yml").expanduser())

    def __init__(self, *args, **kwargs):
        ray.init(address="auto")

        # Collect metadata for the cluster immediately on init
        self._collect_cluster_stats()

        self._installed_packages = []
        self.register_activity()

    def register_activity(self):
        set_last_active_time_to_now()

    def InstallPackages(self, request, context):
        self.register_activity()
        try:
            packages = pickle.loads(request.message)
            logger.info(f"Message received from client to install packages: {packages}")
            for package in packages:
                if isinstance(package, str):
                    pkg = Package.from_string(package)
                elif hasattr(package, "install"):
                    pkg = package
                else:
                    raise ValueError(f"package {package} not recognized")

                if (str(pkg)) in self._installed_packages:
                    continue
                logger.info(f"Installing package: {str(pkg)}")
                pkg.install()
                self._installed_packages.append(str(pkg))

            self.register_activity()
            message = [None, None, None]
        except Exception as e:
            logger.exception(e)
            message = [None, e, traceback.format_exc()]
            self.register_activity()

        return pb2.MessageResponse(message=pickle.dumps(message), received=False)

    def GetObject(self, request, context):
        self.register_activity()
        key, stream_logs = pickle.loads(request.message)
        logger.info(f"Message received from client to get object: {key}")

        logfiles = None
        open_files = None
        ret_obj = None
        returned = False
        while not returned:
            try:
                res = obj_store.get(key, timeout=self.LOGGING_WAIT_TIME)
                logger.info(f"Got object of type {type(res)} back from object store")
                ret_obj = [res, None, None]
                returned = True
                # Don't return yet, go through the loop once more to get any remaining log lines
            except ray.exceptions.GetTimeoutError:
                pass
            except ray.exceptions.TaskCancelledError as e:
                logger.info(f"Attempted to get task {key} that was cancelled.")
                returned = True
                ret_obj = [None, e, traceback.format_exc()]

            if stream_logs:
                if not logfiles:
                    logfiles = obj_store.get_logfiles(key)
                    open_files = [open(i, "r") for i in logfiles]
                    logger.info(f"Streaming logs for {key} from {logfiles}")

                # Grab all the lines written to all the log files since the last time we checked
                ret_lines = []
                for i, f in enumerate(open_files):
                    file_lines = f.readlines()
                    if file_lines:
                        # TODO [DG] handle .out vs .err, and multiple workers
                        # if len(logfiles) > 1:
                        #     ret_lines.append(f"Process {i}:")
                        ret_lines += file_lines
                if ret_lines:
                    yield pb2.MessageResponse(
                        message=pickle.dumps(ret_lines),
                        received=True,
                        output_type=OutputType.STDOUT,
                    )

        if stream_logs:
            # We got the object back from the object store, so we're done (but we went through the loop once
            # more to get any remaining log lines)
            [f.close() for f in open_files]
        yield pb2.MessageResponse(
            message=pickle.dumps(ret_obj), received=True, output_type=OutputType.RESULT
        )

    def PutObject(self, request, context):
        self.register_activity()
        key, obj = pickle.loads(request.message)
        logger.info(f"Message received from client to put object: {key}")
        try:
            obj_store.put(key, obj)
            ret_obj = [key, None, None]
        except Exception as e:
            logger.error(f"Error putting object {key} in object store: {e}")
            ret_obj = [None, e, traceback.format_exc()]
        return pb2.MessageResponse(message=pickle.dumps(ret_obj), received=True)

    def ClearPins(self, request, context):
        self.register_activity()
        pins_to_clear = pickle.loads(request.message)
        logger.info(
            f"Message received from client to clear pins: {pins_to_clear or 'all'}"
        )
        cleared = []
        if pins_to_clear:
            for pin in pins_to_clear:
                remove_pinned_object(pin)
                cleared.append(pin)
        else:
            cleared = list(pinned_keys())
            clear_pinned_memory()

        self.register_activity()
        return pb2.MessageResponse(message=pickle.dumps(cleared), received=True)

    def CancelRun(self, request, context):
        self.register_activity()
        run_keys, force = pickle.loads(request.message)
        if not hasattr(run_keys, "len"):
            run_keys = [run_keys]
        obj_refs = obj_store.get_obj_refs_list(run_keys)
        [
            ray.cancel(obj_ref, force=force, recursive=True)
            for obj_ref in obj_refs
            if isinstance(obj_ref, ray.ObjectRef)
        ]
        return pb2.MessageResponse(
            message=pickle.dumps("Cancelled"),
            received=True,
            output_type=OutputType.RESULT,
        )

    def RunModule(self, request, context):
        self.register_activity()
        # get the function result from the incoming request
        try:
            [relative_path, module_name, fn_name, fn_type, args, kwargs] = pickle.loads(
                request.message
            )

            module_path = (
                str((Path.home() / relative_path).resolve()) if relative_path else None
            )

            if module_name == "notebook":
                fn = fn_name  # Already unpickled above
            else:
                fn = get_fn_by_name(module_name, fn_name, module_path)

            res = call_fn_by_type(fn, fn_type, fn_name, module_path, args, kwargs)
            # [res, None, None] is a silly hack for packaging result alongside exception and traceback
            result = {"message": pickle.dumps([res, None, None]), "received": True}

            self.register_activity()
            return pb2.MessageResponse(**result)
        except Exception as e:
            logger.exception(e)
            message = [None, e, traceback.format_exc()]
            self.register_activity()
            return pb2.MessageResponse(message=pickle.dumps(message), received=False)

    def AddSecrets(self, request, context):
        from runhouse import Secrets

        self.register_activity()
        secrets_to_add: dict = pickle.loads(request.message)
        failed_providers = (
            {}
        )  # Track which providers fail and send them back to the user
        for provider_name, provider_secrets in secrets_to_add.items():
            p = Secrets.builtin_provider_class_from_name(provider_name)
            if p is None:
                error_msg = f"{provider_name} is not a Runhouse builtin provider"
                failed_providers[provider_name] = error_msg
                continue

            # NOTE: For now we are always saving in the provider's default location on the cluster
            credentials_path = p.default_credentials_path()
            try:
                p.save_secrets(provider_secrets, overwrite=True)
            except Exception as e:
                failed_providers[provider_name] = str(e)
                continue

            # update config on the cluster with the default creds path for each provider
            configs.set_nested("secrets", {provider_name: credentials_path})
            logger.info(f"Added secrets for {provider_name} to: {credentials_path}")

        return pb2.MessageResponse(
            message=pickle.dumps(failed_providers),
            received=True,
            output_type=OutputType.RESULT,
        )

    def _collect_cluster_stats(self):
        """Collect cluster metadata and send to Grafana Loki"""
        if configs.get("disable_data_collection") is True:
            return

        cluster_data = self._cluster_status_report()
        sky_data = self._cluster_sky_report()

        self._log_cluster_data(
            {**cluster_data, **sky_data},
            labels={"username": configs.get("username"), "environment": "prod"},
        )

    def _cluster_status_report(self):
        import ray._private.usage.usage_lib as ray_usage_lib
        from ray._private import gcs_utils

        gcs_client = gcs_utils.GcsClient(
            address="127.0.0.1:6379", nums_reconnect_retry=20
        )

        # fields : ['ray_version', 'python_version']
        cluster_metadata = ray_usage_lib.get_cluster_metadata(gcs_client)

        # fields: ['total_num_cpus', 'total_num_gpus', 'total_memory_gb', 'total_object_store_memory_gb']
        cluster_status_report = ray_usage_lib.get_cluster_status_to_report(
            gcs_client
        ).__dict__

        return {**cluster_metadata, **cluster_status_report}

    def _cluster_sky_report(self):
        try:
            from runhouse import Secrets

            sky_ray_data = Secrets.read_yaml_file(self.SKY_YAML)
        except FileNotFoundError:
            # For on prem clusters we won't have sky data
            return {}

        provider = sky_ray_data["provider"]
        node_config = sky_ray_data["available_node_types"].get("ray.head.default", {})

        return {
            "cluster_name": sky_ray_data.get("cluster_name"),
            "region": provider.get("region"),
            "provider": provider.get("module"),
            "instance_type": node_config.get("node_config", {}).get("InstanceType"),
        }

    def _log_cluster_data(self, data: dict, labels: dict):
        from runhouse.rns.api_utils.utils import log_timestamp

        payload = {
            "streams": [
                {"stream": labels, "values": [[str(log_timestamp()), json.dumps(data)]]}
            ]
        }

        payload = json.dumps(payload)
        resp = requests.post(
            f"{configs.get('api_server_url')}/admin/logs", data=json.dumps(payload)
        )

        if resp.status_code == 405:
            # api server not configured to receive grafana logs
            return

        if resp.status_code != 200:
            logger.error(
                f"({resp.status_code}) Failed to send logs to Grafana Loki: {resp.text}"
            )


def serve():
    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=10),
        options=[
            ("grpc.max_send_message_length", UnaryService.MAX_MESSAGE_LENGTH),
            ("grpc.max_receive_message_length", UnaryService.MAX_MESSAGE_LENGTH),
        ],
    )
    pb2_grpc.add_UnaryServicer_to_server(UnaryService(), server)
    server.add_insecure_port("[::]:50052")
    server.start()
    logger.info("Server up and running")
    server.wait_for_termination()


if __name__ == "__main__":
    serve()



from .cluster import Cluster
from .cluster_factory import cluster
from .on_demand_cluster import OnDemandCluster

# TODO KubeRayCluster, AnyscaleRayCluster etc.
#  Other cluster types: Spark? Bare Kubernetes (e.g. EKS, GKS)?
# DaskCluster - should be easy to do with SkyPilot: https://docs.dask.org/en/stable/deploying-ssh.html



import unittest

import runhouse as rh

from runhouse.rns.hardware import cluster, OnDemandCluster


def test_cluster_config():
    rh_cpu = cluster(name="^rh-cpu")
    if not rh_cpu.is_up():
        rh_cpu.up()
    config = rh_cpu.config_for_rns
    cluster2 = OnDemandCluster.from_config(config)
    assert cluster2.address == rh_cpu.address


def test_cluster_sharing():
    c = cluster(name="^rh-cpu").up_if_not().save()
    c.share(
        users=["donny@run.house", "josh@run.house"],
        access_type="write",
        notify_users=False,
    )
    assert True


def test_read_shared_cluster():
    c = cluster(name="/jlewitt1/rh-cpu")
    res = c.run_python(["import numpy", "print(numpy.__version__)"])
    assert res[0][1]


def test_install():
    c = cluster(name="^rh-cpu")
    c.install_packages(
        [
            "./",
            "torch==1.12.1",
            # 'conda:jupyterlab',
            # 'gh:pytorch/vision'
        ]
    )


def test_basic_run():
    # Create temp file where fn's will be stored
    test_cmd = "echo hi"
    hw = cluster(name="^rh-cpu")
    hw.up_if_not()
    res = hw.run(commands=[test_cmd])
    assert "hi" in res[0][1]


def test_restart_grpc():
    hw = cluster(name="^rh-cpu")
    hw.up_if_not()
    codes = hw.restart_grpc_server(resync_rh=False)
    assert codes


def test_same_cluster():
    hw = cluster(name="^rh-cpu")
    hw.up_if_not()

    hw_copy = cluster(name="^rh-cpu")

    def dummy_func(a):
        return a

    func_hw = rh.function(dummy_func).to(hw)
    assert hw.on_same_cluster(func_hw)
    assert hw_copy.on_same_cluster(func_hw)


def test_diff_cluster():
    hw = cluster(name="^rh-cpu")
    hw.up_if_not()

    def dummy_func(a):
        return a

    func_hw = rh.function(dummy_func).to(hw)

    new_hw = cluster(name="diff-cpu")
    assert not new_hw.on_same_cluster(func_hw)


if __name__ == "__main__":
    unittest.main()



import os
from pathlib import Path
from typing import Optional, Tuple

import sky

from runhouse import Secrets


class SkySecrets(Secrets):
    PROVIDER_NAME = "sky"
    PRIVATE_KEY_FILE = os.path.expanduser(
        sky.authentication.PRIVATE_SSH_KEY_PATH
    )  # '~/.ssh/sky-key'
    PUBLIC_KEY_FILE = os.path.expanduser(
        sky.authentication.PUBLIC_SSH_KEY_PATH
    )  # '~/.ssh/sky-key.pub'

    @classmethod
    def default_credentials_path(cls) -> Tuple:
        return cls.PRIVATE_KEY_FILE, cls.PUBLIC_KEY_FILE

    @classmethod
    def has_secrets_file(cls) -> bool:
        secret_files: tuple = cls.default_credentials_path()
        if not secret_files:
            return False
        return all(cls.file_exists(s) for s in secret_files)

    @classmethod
    def read_secrets(cls, from_env: bool = False, file_path: Optional[str] = None):
        if from_env:
            raise NotImplementedError(
                f"Reading secrets from env is not supported for {cls.PROVIDER_NAME}"
            )
        else:
            private_key_file = file_path or cls.PRIVATE_KEY_FILE
            public_key_file = file_path or cls.PUBLIC_KEY_FILE

            private_key = Path(private_key_file).read_text()
            public_key = Path(public_key_file).read_text()

        return {
            "ssh_private_key": private_key,
            "ssh_public_key": public_key,
        }

    @classmethod
    def save_secrets(cls, secrets: dict, overwrite: bool = False):
        private_key_path = cls.PRIVATE_KEY_FILE
        if private_key_path.endswith(".pem"):
            public_key_path = private_key_path.rsplit(".", 1)[0] + ".pub"
        else:
            public_key_path = private_key_path + ".pub"

        cls.check_secrets_for_mismatches(
            secrets_to_save=secrets, secrets_path=public_key_path, overwrite=overwrite
        )

        sky.authentication._save_key_pair(
            private_key_path,
            public_key_path,
            secrets["ssh_private_key"],
            secrets["ssh_public_key"],
        )
        # TODO do we need to register the keys with cloud providers? Probably not, sky does this for us later.
        # backend_utils._add_auth_to_cluster_config(sky.clouds.CLOUD_REGISTRY.from_str(self.provider),
        #                                                   Path(yaml_path).expanduser())
        cls.add_provider_to_rh_config()



import unittest
from pathlib import Path

import runhouse as rh


def setup():
    pass


def test_from_string():
    p = rh.Package.from_string("reqs:~/runhouse")
    assert isinstance(p.install_target, rh.Folder)
    assert p.install_target.path == str(Path.home() / "runhouse")


def test_share_package():
    import shutil

    # Create a local temp folder to install for the package
    tmp_path = Path.cwd().parent / "tmp_package"
    tmp_path.mkdir(parents=True, exist_ok=True)
    for i in range(3):
        output_file = Path(f"{tmp_path}/sample_file_{i}.txt")
        output_file.write_text(f"file{i}")

    p = rh.Package.from_string("local:./tmp_package")
    p.name = "package_to_share"  # shareable resource requires a name

    c = rh.cluster(name="/jlewitt1/rh-cpu")
    p.to_cluster(dest_cluster=c)

    p.share(
        users=["josh@run.house", "donny@run.house"],
        access_type="write",
        notify_users=False,
    )

    shutil.rmtree(tmp_path)

    # Confirm the package's folder is now on the cluster
    status_codes = c.run(commands=["ls tmp_package"])
    assert "sample_file_0.txt" in status_codes[0][1]


def test_share_git_package():
    git_package = rh.GitPackage(
        name="shared_git_package",
        git_url="https://github.com/runhouse/runhouse.git",
        install_method="pip",
        revision="v0.0.1",
    ).save()

    git_package.share(
        users=["donny@run.house", "josh@run.house"],
        access_type="read",
        notify_users=False,
    )
    assert rh.rns_client.exists(name_or_path="shared_git_package")


def test_load_shared_git_package():
    git_package = rh.package(name="@/shared_git_package")
    assert git_package.config_for_rns


if __name__ == "__main__":
    setup()
    unittest.main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: unary.proto
"""Generated protocol buffer code."""
from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database,
)

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="unary.proto",
    package="unary",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x0bunary.proto\x12\x05unary"B\n\x07Message\x12\x0f\n\x07message\x18\x01 \x01(\x0c\x12\x13\n\x0bmodule_name\x18\x02 \x01(\t\x12\x11\n\tfunc_name\x18\x03 \x01(\t"I\n\x0fMessageResponse\x12\x0f\n\x07message\x18\x01 \x01(\x0c\x12\x10\n\x08received\x18\x02 \x01(\x08\x12\x13\n\x0boutput_type\x18\x03 \x01(\t2\x91\x03\n\x05Unary\x12\x35\n\tRunModule\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse"\x00\x12;\n\x0fInstallPackages\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse"\x00\x12\x35\n\tClearPins\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse"\x00\x12\x35\n\tCancelRun\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse"\x00\x12\x35\n\tPutObject\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse"\x00\x12\x36\n\nAddSecrets\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse"\x00\x12\x37\n\tGetObject\x12\x0e.unary.Message\x1a\x16.unary.MessageResponse"\x00\x30\x01\x62\x06proto3',
)


_MESSAGE = _descriptor.Descriptor(
    name="Message",
    full_name="unary.Message",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="message",
            full_name="unary.Message.message",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="module_name",
            full_name="unary.Message.module_name",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="func_name",
            full_name="unary.Message.func_name",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=22,
    serialized_end=88,
)


_MESSAGERESPONSE = _descriptor.Descriptor(
    name="MessageResponse",
    full_name="unary.MessageResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="message",
            full_name="unary.MessageResponse.message",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="received",
            full_name="unary.MessageResponse.received",
            index=1,
            number=2,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="output_type",
            full_name="unary.MessageResponse.output_type",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=90,
    serialized_end=163,
)

DESCRIPTOR.message_types_by_name["Message"] = _MESSAGE
DESCRIPTOR.message_types_by_name["MessageResponse"] = _MESSAGERESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Message = _reflection.GeneratedProtocolMessageType(
    "Message",
    (_message.Message,),
    {
        "DESCRIPTOR": _MESSAGE,
        "__module__": "unary_pb2"
        # @@protoc_insertion_point(class_scope:unary.Message)
    },
)
_sym_db.RegisterMessage(Message)

MessageResponse = _reflection.GeneratedProtocolMessageType(
    "MessageResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _MESSAGERESPONSE,
        "__module__": "unary_pb2"
        # @@protoc_insertion_point(class_scope:unary.MessageResponse)
    },
)
_sym_db.RegisterMessage(MessageResponse)


_UNARY = _descriptor.ServiceDescriptor(
    name="Unary",
    full_name="unary.Unary",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=166,
    serialized_end=567,
    methods=[
        _descriptor.MethodDescriptor(
            name="RunModule",
            full_name="unary.Unary.RunModule",
            index=0,
            containing_service=None,
            input_type=_MESSAGE,
            output_type=_MESSAGERESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="InstallPackages",
            full_name="unary.Unary.InstallPackages",
            index=1,
            containing_service=None,
            input_type=_MESSAGE,
            output_type=_MESSAGERESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="ClearPins",
            full_name="unary.Unary.ClearPins",
            index=2,
            containing_service=None,
            input_type=_MESSAGE,
            output_type=_MESSAGERESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="CancelRun",
            full_name="unary.Unary.CancelRun",
            index=3,
            containing_service=None,
            input_type=_MESSAGE,
            output_type=_MESSAGERESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="PutObject",
            full_name="unary.Unary.PutObject",
            index=4,
            containing_service=None,
            input_type=_MESSAGE,
            output_type=_MESSAGERESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="AddSecrets",
            full_name="unary.Unary.AddSecrets",
            index=5,
            containing_service=None,
            input_type=_MESSAGE,
            output_type=_MESSAGERESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="GetObject",
            full_name="unary.Unary.GetObject",
            index=6,
            containing_service=None,
            input_type=_MESSAGE,
            output_type=_MESSAGERESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_UNARY)

DESCRIPTOR.services_by_name["Unary"] = _UNARY

# @@protoc_insertion_point(module_scope)



import subprocess
import webbrowser
from typing import Optional

import pkg_resources
import typer
from rich.console import Console

from runhouse import cluster, configs
from runhouse.rns import (  # Need to rename it because it conflicts with the login command
    login as login_module,
)

# create an explicit Typer application
app = typer.Typer(add_completion=False)
state = {"verbose": False}

# For printing with typer
console = Console()


@app.command()
def login(token: Optional[str] = typer.Argument(None, help="Your Runhouse API token")):
    """Login to Runhouse. Validates token provided, with options to upload or download stored secrets or config between
    local environment and Runhouse / Vault.
    """
    valid_token: str = login_module.login(token=token, interactive=True, ret_token=True)
    if valid_token:
        webbrowser.open(
            f"{configs.get('api_server_url')}/dashboard?token={valid_token}"
        )
        raise typer.Exit()
    else:
        raise typer.Exit(code=1)


@app.command()
def logout():
    """Logout of Runhouse. Provides options to delete locally configured secrets and local Runhouse configs"""
    login_module.logout(interactive=True)
    raise typer.Exit()


@app.command()
def notebook(
    cluster_name: str, up: bool = typer.Option(False, help="Start the cluster")
):
    """Open a Jupyter notebook on a cluster."""
    c = cluster(name=cluster_name)
    if up:
        c.up_if_not()
    if not c.is_up():
        console.print(
            f"Cluster {cluster_name} is not up. Please run `runhouse notebook {cluster_name} --up`."
        )
        raise typer.Exit(1)
    c.notebook()


@app.command()
def ssh(cluster_name: str, up: bool = typer.Option(False, help="Start the cluster")):
    """SSH into a cluster created elsewhere (so `ssh cluster` doesn't work out of the box) or not yet up."""
    c = cluster(name=cluster_name)
    if up:
        c.up_if_not()
    if not c.is_up():
        console.print(
            f"Cluster {cluster_name} is not up. Please run `runhouse ssh {cluster_name} --up`."
        )
        raise typer.Exit(1)
    subprocess.call(f"ssh {c.name}", shell=True)


@app.command()
def cancel(
    cluster_name: str,
    run_key: str,
    force: Optional[bool] = typer.Option(False, help="Force cancel"),
):
    """Cancel a run on a cluster."""
    c = cluster(name=cluster_name)
    c.cancel(run_key, force=force)


@app.command()
def logs(
    cluster_name: str,
    run_key: str,
    print_results: Optional[bool] = typer.Option(False, help="Print results"),
):
    """Get logs from a run on a cluster."""
    c = cluster(name=cluster_name)
    res = c.get(run_key, stream_logs=True)
    if print_results:
        console.print(res)


def load_cluster(cluster_name: str):
    """Load a cluster from RNS into the local environment, e.g. to be able to ssh."""
    c = cluster(name=cluster_name)
    if not c.address:
        c.update_from_sky_status(dryrun=True)


@app.command()
def restart_grpc(
    cluster_name: str,
    restart_ray: bool = typer.Option(False, help="Restart the Ray runtime"),
    resync_rh: bool = typer.Option(False, help="Resync the Runhouse package"),
):
    """Restart the gRPC server on a cluster."""
    c = cluster(name=cluster_name)
    c.restart_grpc_server(resync_rh=resync_rh, restart_ray=restart_ray)


@app.callback()
def main(verbose: bool = False):
    """
    Runhouse CLI
    """
    if verbose:
        name = "runhouse"
        version = pkg_resources.get_distribution(name).version
        console.print(f"{name}=={version}", style="bold green")
        state["verbose"] = True



import os

import runhouse as rh
import sky
from runhouse.rh_config import configs


def add_secrets_to_vault(headers):
    """Add some test secrets to Vault"""
    # Add real credentials for AWS and SKY to test sky status
    rh.Secrets.put(
        provider="aws",
        secret={
            "access_key": os.getenv("TEST_AWS_ACCESS_KEY"),
            "secret_key": os.getenv("TEST_AWS_SECRET_KEY"),
        },
        headers=headers,
    )
    rh.Secrets.put(
        provider="sky",
        secret={
            "ssh_private_key": os.getenv("TEST_SKY_PRIVATE_KEY"),
            "ssh_public_key": os.getenv("TEST_SKY_PUBLIC_KEY"),
        },
        headers=headers,
    )
    rh.Secrets.put(provider="snowflake", secret={"token": "ABCD1234"}, headers=headers)
    rh.Secrets.put(
        provider="ssh",
        secret={"key-one": "12345", "key-one.pub": "ABCDE"},
        headers=headers,
    )


def test_login_flow_in_new_env():
    token = os.getenv("TEST_TOKEN")
    headers = {"Authorization": f"Bearer {token}"}

    add_secrets_to_vault(headers)

    secrets_in_vault = rh.Secrets.download_into_env(
        headers=headers, save_locally=False, check_enabled=False
    )
    assert secrets_in_vault, "No secrets found in Vault"

    providers_in_vault = list(secrets_in_vault)

    # Login and download secrets stored in Vault into the new env
    rh.login(token=token, download_secrets=True)

    # Once secrets are saved down to their local config, confirm we have sky enabled
    sky.check.check(quiet=True)
    clouds = sky.global_user_state.get_enabled_clouds()
    cloud_names = [str(c).lower() for c in clouds]
    assert "aws" in cloud_names

    enabled_providers = rh.Secrets.enabled_providers(as_str=True)

    for provider in providers_in_vault:
        if provider in enabled_providers:
            assert configs.get("secrets", {}).get(
                provider
            ), f"No credentials path for {provider} stored in .rh config!"

    rh.Secrets.delete_from_vault(providers=["aws", "snowflake", "ssh", "sky"])
    secrets_in_vault = rh.Secrets.download_into_env(
        save_locally=False, check_enabled=False
    )
    assert not secrets_in_vault



import unittest

import runhouse as rh


def test_get_all_secrets_from_vault():
    vault_secrets = rh.Secrets.download_into_env(save_locally=False)
    providers = rh.Secrets.enabled_providers(as_str=True)
    assert set(list(vault_secrets)).issubset(
        providers
    ), "Secrets saved in Vault which are not enabled locally!"


def test_upload_custom_provider_to_vault():
    provider = "sample_provider"
    rh.Secrets.put(provider=provider, secret={"secret_key": "abcdefg"})

    # Retrieve the secret from Vault
    provider_secrets = rh.Secrets.get(provider=provider)
    assert provider_secrets

    rh.Secrets.delete_from_vault(providers=[provider])
    provider_secrets = rh.Secrets.get(provider=provider)
    assert not provider_secrets


def test_upload_aws_to_vault():
    provider = "aws"

    # NOTE: We don't need to provide any secrets here - they will be extracted from the local config file
    rh.Secrets.put(provider=provider)

    # Retrieve the secret from Vault
    provider_secrets = rh.Secrets.get(provider=provider)
    assert provider_secrets

    rh.Secrets.delete_from_vault(providers=[provider])
    provider_secrets = rh.Secrets.get(provider=provider)
    assert not provider_secrets


def test_add_custom_provider():
    import configparser
    import shutil
    from pathlib import Path

    provider = "new_provider"
    creds_dir = Path("~/.config/new_provider").expanduser()
    creds_dir.mkdir(parents=True, exist_ok=True)

    creds_file_path = str(creds_dir / "config")

    parser = configparser.ConfigParser()
    parser.add_section(provider)
    parser.set(section=provider, option="token", value="abcdefg")

    rh.Secrets.save_to_config_file(parser, creds_file_path)
    rh.configs.set_nested("secrets", {provider: creds_file_path})

    # Upload to vault
    rh.Secrets.put(provider, secret={"token": "abcdefg"})
    assert rh.Secrets.get(provider)

    rh.Secrets.delete_from_vault([provider])
    shutil.rmtree(str(creds_dir))
    rh.configs.delete(provider)

    assert not rh.Secrets.get(provider)


def test_upload_all_provider_secrets_to_vault():
    rh.Secrets.extract_and_upload()
    # Download back from Vault
    secrets = rh.Secrets.download_into_env(save_locally=False)
    assert secrets


def test_add_ssh_secrets():
    from runhouse.rns.secrets.ssh_secrets import SSHSecrets

    provider = "ssh"
    # Save to local .ssh directory
    sample_ssh_keys = {"key-one": "12345", "key-one.pub": "ABCDE"}
    SSHSecrets.save_secrets(secrets=sample_ssh_keys, overwrite=True)
    local_secrets: dict = rh.Secrets.load_provider_secrets(providers=[provider])
    assert local_secrets
    assert rh.configs.get("secrets", {}).get(provider)

    # Upload to Vault
    rh.Secrets.put(provider=provider, secret=sample_ssh_keys)
    vault_secrets = rh.Secrets.get(provider=provider)
    assert vault_secrets

    # Delete from Vault & local configs
    rh.configs.delete(provider)
    rh.Secrets.delete_from_vault([provider])
    for f_name, _ in sample_ssh_keys.items():
        ssh_key_path = f"{SSHSecrets.default_credentials_path()}/{f_name}"
        rh.Secrets.delete_secrets_file(file_path=ssh_key_path)

    assert not rh.Secrets.get(provider)
    assert not rh.configs.get("secrets", {}).get(provider)


def test_add_github_secrets():
    from runhouse.rns.secrets.github_secrets import GitHubSecrets

    provider = "github"

    vault_secrets = rh.Secrets.get(provider=provider)
    if not vault_secrets:
        # Create mock tokens and test uploading / downloading / deleting
        # Upload to Vault
        sample_gh_token = {"token": "12345"}
        rh.Secrets.put(provider=provider, secret=sample_gh_token)

        # save to local config file
        GitHubSecrets.save_secrets(secrets=sample_gh_token, overwrite=True)
        assert rh.configs.rh.configs.get("secrets", {}).get(provider)
        assert rh.Secrets.get(provider)

        # Delete from Vault & local config
        rh.Secrets.delete_from_vault(providers=[provider])
        rh.Secrets.delete_from_local_env(providers=[provider])

        assert not rh.Secrets.get(provider)
        assert not rh.configs.get("secrets", {}).get(provider)

    else:
        # Load from local config so as not to overwrite existing GitHub secrets
        local_secrets: dict = rh.Secrets.load_provider_secrets(providers=[provider])
        assert local_secrets
        assert rh.configs.get("secrets", {}).get(provider)


def test_sending_secrets_to_cluster():
    cluster = rh.cluster(name="^rh-cpu").up_if_not()
    enabled_providers: list = rh.Secrets.enabled_providers()

    cluster.send_secrets(providers=enabled_providers)

    # Confirm the secrets now exist on the cluster
    for provider_cls in enabled_providers:
        provider_name = provider_cls.PROVIDER_NAME
        commands = [
            f"from runhouse.rns.secrets.{provider_name}_secrets import {str(provider_cls)}",
            f"print({str(provider_cls)}.has_secrets_file())",
        ]
        status_codes: list = cluster.run_python(commands)
        if "False" in status_codes[0][1]:
            assert False, f"No credentials file found on cluster for {provider_name}"

    assert True


@unittest.skip("for manually debugging full login flow")
def test_login_manual():
    rh.login(
        token="...",
        download_config=False,
        download_secrets=True,
        upload_secrets=False,
        upload_config=False,
    )


@unittest.skip("This test overrides local rh token if done incorrectly")
# Running this unit test will override local rh config token to "..."
# Was meant to be run by manually inputting token, but need a better way to test
def test_login():
    # TODO [DG] create a mock account and test this properly in CI
    token = "..."

    rh.login(
        token=token,
        download_config=False,
        download_secrets=False,
        upload_secrets=False,
        upload_config=False,
    )

    rh.login(
        token=token,
        download_config=False,
        download_secrets=False,
        upload_secrets=True,
        upload_config=True,
    )

    rh.login(
        token=token,
        download_config=True,
        download_secrets=True,
        upload_secrets=False,
        upload_config=False,
    )

    rh.login(
        token=token,
        download_config=True,
        download_secrets=True,
        upload_secrets=True,
        upload_config=True,
    )

    assert rh.rns_client.default_folder == "/..."


def test_logout():
    enabled_providers = rh.Secrets.enabled_providers(as_str=True)
    current_config: dict = rh.configs.load_defaults_from_file()

    rh.logout(delete_loaded_secrets=True, delete_rh_config_file=True)

    for provider_name in enabled_providers:
        p = rh.Secrets.builtin_provider_class_from_name(provider_name)
        assert not p.has_secrets_file()
        assert not rh.configs.get(provider_name)

    assert not rh.configs.load_defaults_from_file()

    # Restore the config and secrets to its pre-logout state
    rh.configs.save_defaults(defaults=current_config)
    secrets = rh.Secrets.download_into_env(
        providers=enabled_providers, save_locally=True
    )
    assert secrets


if __name__ == "__main__":
    unittest.main()



import shutil
import unittest
from pathlib import Path

import datasets
import pandas as pd
import pyarrow as pa
import ray.data

import runhouse as rh
from runhouse import Folder

TEMP_LOCAL_FOLDER = Path("~/.rh/temp").expanduser()
BUCKET_NAME = "runhouse-table"
NUM_PARTITIONS = 10


def setup():
    from runhouse.rns.api_utils.utils import create_s3_bucket

    create_s3_bucket(BUCKET_NAME)


def delete_local_folder(path):
    shutil.rmtree(path)


def tokenize_function(examples):
    from transformers import AutoTokenizer

    tokenizer = AutoTokenizer.from_pretrained("bert-base-cased")
    return tokenizer(examples["text"], padding="max_length", truncation=True)


def load_sample_data(data_type):
    if data_type == "huggingface":
        from datasets import load_dataset

        dataset = load_dataset("yelp_review_full", split="train[:1%]")
        return dataset

    elif data_type == "pyarrow":
        df = pd.DataFrame(
            {
                "int": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                "str": ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"],
            }
        )
        arrow_table = pa.Table.from_pandas(df)
        return arrow_table

    elif data_type == "cudf":
        import cudf

        gdf = cudf.DataFrame(
            {"id": [1, 2, 3, 4, 5, 6], "grade": ["a", "b", "b", "a", "a", "e"]}
        )
        return gdf

    elif data_type == "pandas":
        df = pd.DataFrame(
            {"id": [1, 2, 3, 4, 5, 6], "grade": ["a", "b", "b", "a", "a", "e"]}
        )
        return df

    elif data_type == "dask":
        import dask.dataframe as dd

        index = pd.date_range("2021-09-01", periods=2400, freq="1H")
        df = pd.DataFrame({"a": range(2400), "b": list("abcaddbe" * 300)}, index=index)
        ddf = dd.from_pandas(df, npartitions=NUM_PARTITIONS)
        return ddf

    elif data_type == "ray":
        import ray

        ds = ray.data.range(10000)
        return ds

    else:
        raise Exception(f"Unsupported data type {data_type}")


# -----------------------------------------------
# ----------------- Local tests -----------------
# -----------------------------------------------


def test_create_and_reload_file_locally():
    local_path = Path.cwd() / "table_tests/local_test_table"
    local_path.mkdir(parents=True, exist_ok=True)

    Path(local_path).mkdir(parents=True, exist_ok=True)

    orig_data = pd.DataFrame({"my_col": list(range(50))})
    name = "~/my_local_test_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=str(local_path),
            system="file",
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = reloaded_table.data

    assert reloaded_data.to_pandas().equals(orig_data)

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pd.DataFrame)
        assert batch["my_col"].tolist() == list(range(idx * 10, (idx + 1) * 10))

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_pandas_locally():
    orig_data = load_sample_data("pandas")
    name = "~/my_test_local_pandas_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            system="file",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = reloaded_table.data

    assert orig_data.equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pd.DataFrame)
        assert batch["id"].tolist() == list(range(1, 7))

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_pyarrow_locally():
    orig_data = load_sample_data("pyarrow")
    name = "~/my_test_local_pyarrow_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            system="file",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = reloaded_table.data

    assert orig_data.to_pandas().equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pa.Table)
        assert batch["int"].to_pylist() == list(range(1, 11))

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_ray_locally():
    orig_data = load_sample_data("ray")
    name = "~/my_test_local_ray_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            system="file",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = reloaded_table.data

    assert orig_data.to_pandas().equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pa.Table)
        assert batch["value"].to_pylist() == list(range(idx * 10, (idx + 1) * 10))

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_huggingface_locally():
    orig_data = load_sample_data("huggingface")
    name = "~/my_test_local_huggingface_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            system="file",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = reloaded_table.data

    assert orig_data.to_pandas().equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10, as_dict=False)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["label", "text"]
        assert batch.shape == (10, 2)

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_dask_locally():
    orig_data = load_sample_data("dask")
    name = "~/my_test_local_dask_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path="table_tests/dask_test_table",
            system="file",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: "dask.dataframe.core.DataFrame" = reloaded_table.data.to_dask()
    assert reloaded_data.columns.to_list() == ["a", "b"]

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["a", "b"]
        assert batch.shape == (10, 2)

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


# --------------------------------------------
# ----------------- S3 tests -----------------
# --------------------------------------------
def test_create_and_reload_pyarrow_data_from_s3():
    orig_data = load_sample_data(data_type="pyarrow")
    name = "@/my_test_pyarrow_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/pyarrow_df",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert reloaded_data.to_pandas().equals(orig_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["int", "str"]

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_pandas_data_from_s3():
    orig_data = load_sample_data(data_type="pandas")
    name = "@/my_test_pandas_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/pandas_df",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert orig_data.equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pd.DataFrame)
        assert batch["id"].tolist() == list(range(1, 7))

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_huggingface_data_from_s3():
    orig_data: datasets.Dataset = load_sample_data(data_type="huggingface")
    name = "@/my_test_hf_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/huggingface_data",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)

    # Stream in as huggingface dataset
    batches = reloaded_table.stream(batch_size=10, as_dict=False)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["label", "text"]
        assert batch.shape == (10, 2)

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_dask_data_from_s3():
    orig_data = load_sample_data(data_type="dask")
    name = "@/my_test_dask_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/dask",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: "dask.dataframe.core.DataFrame" = reloaded_table.data.to_dask()
    assert reloaded_data.columns.to_list() == ["a", "b"]

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["a", "b"]
        assert batch.shape == (10, 2)

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_ray_data_from_s3():
    orig_data = load_sample_data(data_type="ray")
    name = "@/my_test_ray_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/ray_data",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert reloaded_data.to_pandas().equals(orig_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pa.Table)
        assert batch["value"].to_pylist() == list(range(idx * 10, (idx + 1) * 10))

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


# ----------------- Iter -----------------
def test_load_pandas_data_as_iter():
    orig_data = load_sample_data(data_type="pandas")
    name = "@/my_test_pandas_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/pandas",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = next(iter(reloaded_table))

    assert isinstance(reloaded_data, pd.Series)
    assert reloaded_data.to_dict() == {"id": 1, "grade": "a"}

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_load_pyarrow_data_as_iter():
    orig_data = load_sample_data(data_type="pyarrow")
    name = "@/my_test_pyarrow_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/pyarrow-data",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: pa.ChunkedArray = next(iter(reloaded_table))

    assert isinstance(reloaded_data, pa.ChunkedArray)
    assert reloaded_data.to_pylist() == list(range(1, 11))

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_load_huggingface_data_as_iter():
    orig_data = load_sample_data(data_type="huggingface")
    name = "@/my_test_huggingface_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/huggingface-dataset",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: pa.ChunkedArray = next(iter(reloaded_table))
    assert isinstance(reloaded_data, pa.ChunkedArray)

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


# ----------------- Shuffling -----------------
def test_shuffling_pyarrow_data_from_s3():
    orig_data = load_sample_data(data_type="pyarrow")
    name = "@/my_test_shuffled_pyarrow_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/pyarrow",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    batches = reloaded_table.stream(
        batch_size=10, shuffle_seed=42, shuffle_buffer_size=10
    )
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pa.Table)
        assert orig_data.columns[0].to_pylist() != batch.columns[0].to_pylist()
        assert orig_data.columns[1].to_pylist() != batch.columns[1].to_pylist()

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


# -------------------------------------------------
# ----------------- Cluster tests -----------------
# -------------------------------------------------
def test_create_and_reload_pandas_data_from_cluster():
    cluster = rh.cluster(name="^rh-cpu").up_if_not()

    # Make sure the destination folder for the data exists on the cluster
    data_path_on_cluster = f"{Folder.DEFAULT_CACHE_FOLDER}/pandas-data"
    cluster.run([f"mkdir -p {data_path_on_cluster}"])

    orig_data = load_sample_data(data_type="pandas")
    name = "@/my_test_pandas_table"
    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=data_path_on_cluster,
            system=cluster,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)

    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert orig_data.equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pd.DataFrame)
        assert batch["id"].tolist() == list(range(1, 7))

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_ray_data_from_cluster():
    cluster = rh.cluster("^rh-cpu").up_if_not().save()

    data_path_on_cluster = f"{Folder.DEFAULT_CACHE_FOLDER}/ray-data"
    cluster.run([f"mkdir -p {data_path_on_cluster}"])

    orig_data = load_sample_data(data_type="ray")
    name = "@/my_test_ray_cluster_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=data_path_on_cluster,
            system=cluster,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert orig_data.to_pandas().equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert isinstance(batch, pa.Table)
        assert batch["value"].to_pylist() == list(range(idx * 10, (idx + 1) * 10))

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_pyarrow_data_from_cluster():
    cluster = rh.cluster("^rh-cpu").up_if_not().save()

    data_path_on_cluster = f"{Folder.DEFAULT_CACHE_FOLDER}/pyarrow-data"
    cluster.run([f"mkdir -p {data_path_on_cluster}"])

    orig_data = load_sample_data(data_type="pyarrow")
    name = "@/my_test_pyarrow_cluster_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=data_path_on_cluster,
            system=cluster,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert orig_data.to_pandas().equals(reloaded_data.to_pandas())

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["int", "str"]

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_huggingface_data_from_cluster():
    cluster = rh.cluster("^rh-cpu").up_if_not().save()

    data_path_on_cluster = f"{Folder.DEFAULT_CACHE_FOLDER}/hf-data"
    cluster.run([f"mkdir -p {data_path_on_cluster}"])

    orig_data = load_sample_data(data_type="huggingface")
    name = "@/my_test_hf_cluster_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=data_path_on_cluster,
            system=cluster,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = reloaded_table.data
    assert orig_data.to_pandas().equals(reloaded_data.to_pandas())

    # Stream in as huggingface dataset
    batches = reloaded_table.stream(batch_size=10, as_dict=False)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["label", "text"]
        assert batch.shape == (10, 2)

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_reload_dask_data_from_cluster():
    cluster = rh.cluster("^rh-cpu").up_if_not().save()

    data_path_on_cluster = f"{Folder.DEFAULT_CACHE_FOLDER}/dask-data"
    cluster.run([f"mkdir -p {data_path_on_cluster}"])

    orig_data = load_sample_data(data_type="dask")
    name = "@/my_test_dask_cluster_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=data_path_on_cluster,
            system=cluster,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: "dask.dataframe.core.DataFrame" = reloaded_table.data.to_dask()
    assert reloaded_data.columns.to_list() == ["a", "b"]

    batches = reloaded_table.stream(batch_size=10)
    for idx, batch in enumerate(batches):
        assert batch.column_names == ["a", "b"]
        assert batch.shape == (10, 2)

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_to_cluster_attr():
    local_path = Path.cwd() / "table_tests/local_test_table"
    local_path.mkdir(parents=True, exist_ok=True)

    Path(local_path).mkdir(parents=True, exist_ok=True)

    orig_data = pd.DataFrame({"my_col": list(range(50))})
    name = "~/my_local_test_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=str(local_path),
            system="file",
        )
        .write()
        .save()
    )

    cluster = rh.cluster("^rh-cpu").up_if_not().save()
    cluster_table = my_table.to(system=cluster)

    assert isinstance(cluster_table.system, rh.Cluster)
    assert cluster_table._folder._fs_str == "ssh"

    data = cluster_table.data
    assert data.to_pandas().equals(orig_data)

    del orig_data
    del my_table

    cluster_table.delete_configs()
    cluster_table.delete_in_system()
    assert not cluster_table.exists_in_system()


# -------------------------------------------------
# ----------------- Fetching tests -----------------
# -------------------------------------------------
def test_create_and_fetch_pyarrow_data_from_s3():
    orig_data = load_sample_data(data_type="pyarrow")
    name = "@/my_test_fetch_pyarrow_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/pyarrow",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: pa.Table = reloaded_table.fetch()
    assert orig_data == reloaded_data

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_fetch_pandas_data_from_s3():
    orig_data = load_sample_data(data_type="pandas")
    name = "@/my_test_fetch_pandas_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/pandas",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: pd.DataFrame = reloaded_table.fetch()
    assert orig_data.equals(reloaded_data)

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_fetch_huggingface_data_from_s3():
    orig_data = load_sample_data(data_type="huggingface")
    name = "@/my_test_fetch_huggingface_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/huggingface",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: datasets.Dataset = reloaded_table.fetch()
    assert orig_data.description == reloaded_data.description

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_fetch_ray_data_from_s3():
    orig_data = load_sample_data(data_type="ray")
    name = "@/my_test_fetch_ray_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/ray",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: ray.data.Dataset = reloaded_table.fetch()
    assert orig_data.to_pandas().equals(reloaded_data.to_pandas())

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


def test_create_and_fetch_dask_data_from_s3():
    orig_data = load_sample_data(data_type="dask")
    name = "@/my_test_fetch_dask_table"
    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            path=f"/{BUCKET_NAME}/dask",
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    reloaded_table = rh.Table.from_name(name)
    reloaded_data: "dask.dataframe.core.DataFrame" = reloaded_table.fetch()
    assert orig_data.npartitions == reloaded_data.npartitions

    del orig_data
    del my_table

    reloaded_table.delete_configs()

    reloaded_table.delete_in_system()
    assert not reloaded_table.exists_in_system()


# -------------------------------------------------
# ----------------- Table Sharing tests -----------------
# -------------------------------------------------
def test_sharing_table():
    orig_data = load_sample_data(data_type="pandas")
    name = "shared_pandas_table"

    my_table = (
        rh.table(
            data=orig_data,
            name=name,
            system="s3",
            mkdir=True,
        )
        .write()
        .save()
    )

    my_table.share(
        users=["donny@run.house", "josh@run.house"],
        access_type="write",
        notify_users=False,
    )

    assert my_table.exists_in_system()


def test_read_shared_table():
    my_table = rh.Table.from_name(name="@/shared_pandas_table")
    df: pd.DataFrame = my_table.fetch()
    assert not df.empty


if __name__ == "__main__":
    setup()
    unittest.main()



import json
import logging
import os
import pkgutil
import shutil
from pathlib import Path
from typing import Optional

import requests

from runhouse.rns.api_utils.resource_access import ResourceAccess
from runhouse.rns.api_utils.utils import (
    generate_uuid,
    load_resp_content,
    read_resp_data,
    remove_null_values_from_dict,
)

logger = logging.getLogger(__name__)


class RNSClient:
    """Manage a particular resource with the runhouse database"""

    CORE_RNS_FIELDS = ["name", "resource_type", "folder", "users", "groups"]
    # RH_BUILTINS_FOLDER = '/builtins'
    DEFAULT_FS = "file"

    def __init__(self, configs) -> None:
        self._configs = configs
        self._prev_folders = []

        self.rh_directory = str(Path(self.locate_working_dir()) / "rh")
        self.rh_builtins_directory = str(
            Path(pkgutil.get_loader("runhouse").path).parent / "builtins"
        )

        # TODO allow users to register other base folders
        # Register all the directories in rh folder as rns base folders
        rns_base_folders = (
            [
                str(subdir)
                for subdir in Path(self.rh_directory).iterdir()
                if subdir.is_dir()
            ]
            if Path(self.rh_directory).exists()
            else []
        )
        rns_base_folders.append(
            str(Path(pkgutil.get_loader("runhouse").path).parent / "builtins")
        )
        self._index_base_folders(rns_base_folders)
        self._current_folder = None

        self.refresh_defaults()

    # TODO [DG] move the below into Defaults() so they never need to be refreshed?
    def refresh_defaults(self):
        use_local_configs = (
            ["local"] if self._configs.get("use_local_configs", True) else []
        )
        use_rns = (
            ["rns"]
            if self._configs.get("use_rns", self._configs.get("token", False))
            else []
        )

        self.save_to = use_local_configs + use_rns
        self.load_from = use_local_configs + use_rns

        if self.token is None:
            self.save_to.pop(
                self.save_to.index("rns")
            ) if "rns" in self.save_to else self.save_to
            self.load_from.pop(
                self.load_from.index("rns")
            ) if "rns" in self.load_from else self.load_from
            logger.info(
                "No auth token provided, so not using RNS API to save and load configs"
            )

    @classmethod
    def find_parent_with_file(cls, dir_path, file):
        if Path(dir_path) == Path.home() or dir_path == Path("/"):
            return None
        if Path(dir_path, file).exists():
            return str(dir_path)
        else:
            return cls.find_parent_with_file(Path(dir_path).parent, file)

    @classmethod
    def locate_working_dir(cls, cwd=os.getcwd()):
        # Search for working_dir by looking up directory tree, in the following order:
        # 1. Upward directory with rh/ subdirectory
        # 2. Root git directory
        # 3. Upward directory with requirements.txt
        # 4. User's cwd

        for search_target in [
            "rh",
            ".git",
            "requirements.txt",
            "setup.py",
            "pyproject.toml",
        ]:
            dir_with_target = cls.find_parent_with_file(cwd, search_target)
            if dir_with_target is not None:
                return dir_with_target
        else:
            return cwd

    @property
    def default_folder(self):
        folder = self._configs.get("default_folder", None)
        if folder in [None, "~"] and self._configs.get("username"):
            folder = "/" + self._configs.get("username")
            self._configs.set("default_folder", folder)
        return folder

    @property
    def current_folder(self):
        if not self._current_folder:
            self._current_folder = self.default_folder
        return self._current_folder

    @current_folder.setter
    def current_folder(self, value):
        self._current_folder = value

    @property
    def token(self):
        return self._configs.get("token", None)

    @property
    def api_server_url(self):
        return self._configs.get("api_server_url", None)

    def _index_base_folders(self, lst):
        self.rns_base_folders = {}
        for folder in lst:
            config = self._load_config_from_local(path=folder)
            rns_path = str(Path(self.default_folder) / Path(folder).name)
            if config:
                rns_path = config.get("rns_address")
            self.rns_base_folders[rns_path] = folder

    @staticmethod
    def resource_uri(name):
        """URI used when querying the RNS server"""
        from runhouse.rns.top_level_rns_fns import resolve_rns_path

        rns_address = resolve_rns_path(name)
        return RNSClient.format_rns_address(rns_address)

    @staticmethod
    def format_rns_address(rns_address: str):
        if rns_address.startswith("/"):
            rns_address = rns_address[1:]
        return rns_address.replace("/", ":")

    @staticmethod
    def local_to_remote_address(rns_address):
        return rns_address.replace("~", "@")

    def remote_to_local_address(self, rns_address):
        return rns_address.replace(self.default_folder, "~")

    @property
    def request_headers(self):
        return self._configs.request_headers

    def resource_request_payload(self, payload) -> dict:
        payload = remove_null_values_from_dict(payload)
        data = {}
        for k, v in payload.copy().items():
            if k not in self.CORE_RNS_FIELDS:
                data[k] = v
                # if adding to data field remove as standalone field
                del payload[k]
        payload["data"] = data
        return payload

    def grant_resource_access(
        self,
        rns_address: str,
        user_emails: list,
        access_type: ResourceAccess,
        notify_users: bool,
    ):
        resource_uri = self.resource_uri(rns_address)
        headers = self.request_headers
        access_payload = {
            "users": user_emails,
            "access_type": access_type,
            "notify_users": notify_users,
        }
        uri = "resource/" + resource_uri
        resp = requests.put(
            f"{self.api_server_url}/{uri}/users/access",
            data=json.dumps(access_payload),
            headers=headers,
        )
        if resp.status_code != 200:
            raise Exception(f"Failed to grant access: {load_resp_content(resp)}")

        resp_data: dict = read_resp_data(resp)
        added_users: dict = resp_data.get("added_users", {})
        new_users: dict = resp_data.get("new_users", {})

        return added_users, new_users

    def load_config(
        self,
        name,
    ) -> dict:
        if not name:
            return {}

        rns_address = self.resolve_rns_path(name)

        if rns_address[0] in ["~", "^"]:
            config = self._load_config_from_local(rns_address)
            if config:
                return config

        if rns_address.startswith("/"):
            resource_uri = self.resource_uri(name)
            logger.info(f"Attempting to load config for {rns_address} from RNS.")
            uri = "resource/" + resource_uri
            resp = requests.get(
                f"{self.api_server_url}/{uri}", headers=self.request_headers
            )
            if resp.status_code != 200:
                logger.info(f"No config found in RNS: {load_resp_content(resp)}")
                # No config found, so return empty config
                return {}

            config: dict = read_resp_data(resp)
            if config.get("data", None):
                config.update(config["data"])
                del config["data"]
            return config
        return {}

    def _load_config_from_local(self, rns_address=None, path=None) -> Optional[dict]:
        """Load config from local file"""
        # TODO should we handle remote filessytems, or throw an error if system != 'file'?
        if not path:
            path = self.locate(rns_address, resolve_path=False)
            if not path:
                return None
        config_path = Path(path) / "config.json"
        if not config_path.exists():
            return None

        logger.info(f"Loading config from local file {config_path}")
        with open(config_path, "r") as f:
            try:
                config = json.load(f)
            except json.decoder.JSONDecodeError as e:
                logger.error(f"Error loading config from {config_path}: {e}")
                return None
        if rns_address:
            config["name"] = rns_address
        return config

    def get_rns_address_for_local_path(self, local_path):
        """Get RNS address for local path"""
        try:
            rel_path = str(Path(local_path).relative_to(self.rh_directory))
            return "~/" + rel_path
        except ValueError:
            return None

    def save_config(self, resource, overwrite: bool = True):
        """Register the resource, saving it to local config folder and/or RNS config store. Uses the resource's
        `self.config_for_rns` to generate the dict to save."""
        rns_address = resource.rns_address
        config = resource.config_for_rns

        if not overwrite and self.exists(rns_address):
            raise ValueError(
                f"Resource {rns_address} already exists and overwrite is False."
            )

        config["name"] = rns_address

        if rns_address is None:
            return

        if rns_address[0] in ["~", "^"]:
            self._save_config_to_local(config, rns_address)

        if rns_address.startswith("/"):
            self._save_config_in_rns(config, rns_address)

    def _save_config_to_local(self, config: dict, rns_address: str):
        if not rns_address:
            raise ValueError("Cannot save resource without rns address or path.")
        resource_dir = Path(self.locate(rns_address, resolve_path=False))
        resource_dir.mkdir(parents=True, exist_ok=True)
        config_path = resource_dir / "config.json"
        with open(config_path, "w") as f:
            json.dump(config, f, indent=4)
        logger.info(f"Saving config for {rns_address} to: {config_path}")

    def _save_config_in_rns(self, config, resource_name):
        """Update or create resource config in database"""
        logger.info(f"Saving config to RNS: {config}")

        resource_uri = self.resource_uri(resource_name)
        uri = f"resource/{resource_uri}"

        payload = self.resource_request_payload(config)
        headers = self.request_headers
        resp = requests.put(
            f"{self.api_server_url}/{uri}", data=json.dumps(payload), headers=headers
        )
        if resp.status_code == 200:
            logger.info(f"Config updated in RNS for Runhouse URI <{uri}>")
        elif resp.status_code == 422:  # No changes made to existing Resource
            logger.info(f"Config for {uri} has not changed, nothing to update")
        elif resp.status_code == 404:  # Resource not found
            logger.info(f"Saving new resource in RNS for Runhouse URI <{uri}>")
            # Resource does not yet exist, in which case we need to create from scratch
            resp = requests.post(
                f"{self.api_server_url}/resource",
                data=json.dumps(payload),
                headers=headers,
            )
            if resp.status_code != 200:
                raise Exception(
                    f"Failed to create new resource in RNS: {load_resp_content(resp)}"
                )
        else:
            raise Exception(
                f"Failed to save resource <{uri}> in RNS: {load_resp_content(resp)}"
            )

    def delete_configs(
        self,
        resource,
    ):
        rns_address = (
            resource.rns_address
            if hasattr(resource, "rns_address")
            else self.resolve_rns_path(resource)
        )
        if rns_address is None:
            logger.warning("No rns address exists for resource")
            return

        if rns_address[0] in ["~", "^"]:
            path = self.locate(rns_address, resolve_path=False)
            if path and Path(path).exists():
                shutil.rmtree(path)
            else:
                logger.info(
                    f"Cannot delete resource {rns_address}, could not find the local config."
                )

        if rns_address.startswith("/"):
            resource_uri = self.resource_uri(rns_address)
            uri = "resource/" + resource_uri
            resp = requests.delete(
                f"{self.api_server_url}/{uri}", headers=self.request_headers
            )
            if resp.status_code != 200:
                logger.error(f"Failed to delete_configs <{uri}>")
            else:
                logger.info(f"Successfully deleted <{uri}>")

    def resolve_rns_data_resource_name(self, name: str):
        """If no name is explicitly provided for the data resource, we need to create one based on the relevant
        rns path. If name is None, return a hex uuid.
        For example: my_blob -> jlewitt1/my_blob"
        """
        if name is None:
            return generate_uuid()
        rns_path = self.resolve_rns_path(name)
        if rns_path.startswith("~"):
            return rns_path[2:]
        # For the purposes of building the path to the underlying data resource we don't need the slash
        return rns_path.lstrip("/")

    #########################
    # Folder Operations
    #########################

    def resolve_rns_path(self, path: str):
        if path == ".":
            return self.current_folder
        if path.startswith("./"):
            return self.current_folder + "/" + path[2:]
        # if path == '~':
        #     return '/rh'
        # if path.startswith('~/'):
        #     return '/rh/' + path[2:]
        # TODO break out paths for remote rns?
        if path == "@":
            return self.default_folder
        if path.startswith("@/"):
            return self.default_folder + "/" + path[2:]
        # if path == '^':
        #     return self.RH_BUILTINS_FOLDER
        # if path.startswith('^'):
        #     return self.RH_BUILTINS_FOLDER + '/' + path[1:]
        if path[0] not in ["/", "~", "^"]:
            return self.current_folder + "/" + path
        return path

    @staticmethod
    def split_rns_name_and_path(path: str):
        return Path(path).name, str(Path(path).parent)

    def exists(
        self,
        name_or_path,
        resource_type: str = None,
    ):
        config = self.load_config(name_or_path)
        if not config:
            return False
        if resource_type:
            return config.get("resource_type") == resource_type
        return True

    def locate(
        self,
        name,
        resolve_path=True,
    ):
        """Return the path for a resource."""
        # First check if name is in current folder

        if name == "/":
            return None

        if resolve_path:
            name = self.resolve_rns_path(name)

        if name.startswith("~"):
            return name.replace("~", self.rh_directory)

        if name.startswith("^"):
            return name.replace("^", self.rh_builtins_directory + "/")

        # TODO [DG] see if this breaks anything, also make it traverse the various rns folders to find the resource
        # if name.startswith('/'):
        #     if self.exists(name):
        #         return self.resource_uri(name)

        return None

    def set_folder(self, path: str, create=False):
        from runhouse.rns.folders.folder import Folder, folder

        if isinstance(path, Folder):
            abs_path = path.rns_address
        else:
            abs_path = self.resolve_rns_path(path)
            if abs_path in ["~", "~/"]:
                create = False
            if create:
                folder(path=path, dryrun=False)

        self._prev_folders += [self.current_folder]
        self.current_folder = abs_path

    def unset_folder(self):
        """Sort of like `cd -`, but with a full stack of the previous folder's set. Resets the
        current_folder to the previous one on the stack, the current_folder right before the
        current one was set."""
        if len(self._prev_folders) == 0:
            # TODO should we be raising an error here?
            return
        self.current_folder = self._prev_folders.pop(-1)

    def contents(self, name_or_path, full_paths):
        from runhouse.rns.folders.folder import folder

        folder_url = self.locate(name_or_path)
        return folder(name=name_or_path, path=folder_url).resources(
            full_paths=full_paths
        )



import configparser
import os
from typing import Optional

from runhouse import Secrets


class AzureSecrets(Secrets):
    PROVIDER_NAME = "azure"
    CREDENTIALS_FILE = os.path.expanduser("~/.azure/clouds.config")

    @classmethod
    def read_secrets(cls, from_env: bool = False, file_path: Optional[str] = None):
        if from_env:
            subscription_id = os.getenv("AZURE_SUBSCRIPTION_ID")
            if not subscription_id:
                raise Exception(
                    f"AZURE_SUBSCRIPTION_ID must is not set for {cls.PROVIDER_NAME}"
                )
        else:
            creds_file = file_path or cls.default_credentials_path()
            config = cls.read_config_file(creds_file)
            subscription_id = config["AzureCloud"]["subscription"]

        return {"subscription_id": subscription_id}

    @classmethod
    def save_secrets(cls, secrets: dict, overwrite: bool = False):

        dest_path = cls.default_credentials_path()
        cls.check_secrets_for_mismatches(
            secrets_to_save=secrets, secrets_path=dest_path, overwrite=overwrite
        )

        parser = configparser.ConfigParser()
        section_name = "AzureCloud"
        parser.add_section(section_name)
        parser.set(
            section=section_name,
            option="subscription",
            value=secrets["subscription_id"],
        )

        cls.save_to_config_file(parser, dest_path)
        cls.add_provider_to_rh_config()



import logging
import time
import unittest

import runhouse as rh
from tqdm.auto import tqdm  # progress bar

from tests.test_function import multiproc_torch_sum

TEMP_FILE = "my_file.txt"
TEMP_FOLDER = "~/runhouse-tests"

logger = logging.getLogger(__name__)


def get_test_cluster():
    return rh.cluster(name="^rh-cpu").up_if_not()


def do_printing_and_logging():
    for i in range(6):
        # Wait to make sure we're actually streaming
        time.sleep(1)
        print(f"Hello from the cluster! {i}")
        logger.info(f"Hello from the cluster logs! {i}")
    return list(range(50))


def do_tqdm_printing_and_logging(steps=6):
    progress_bar = tqdm(range(steps))
    for i in range(steps):
        # Wait to make sure we're actually streaming
        time.sleep(0.1)
        progress_bar.update(1)
    return list(range(50))


def test_get_from_cluster():
    cluster = get_test_cluster()
    print_fn = rh.function(fn=do_printing_and_logging, system=cluster)
    key = print_fn.remote()
    assert isinstance(key, str)
    res = cluster.get(key, stream_logs=True)
    assert res == list(range(50))


def test_put_and_get_on_cluster():
    test_list = list(range(5, 50, 2)) + ["a string"]
    cluster = get_test_cluster()
    cluster.put("my_list", test_list)
    ret = cluster.get("my_list")
    assert all(a == b for (a, b) in zip(ret, test_list))


def test_stream_logs():
    cluster = get_test_cluster()
    print_fn = rh.function(fn=do_printing_and_logging, system=cluster)
    res = print_fn(stream_logs=True)
    # TODO [DG] assert that the logs are streamed
    assert res == list(range(50))


def test_multiprocessing_streaming():
    cluster = get_test_cluster()
    re_fn = rh.function(
        multiproc_torch_sum, system=cluster, reqs=["./", "torch==1.12.1"]
    )
    summands = list(zip(range(5), range(4, 9)))
    res = re_fn(summands, stream_logs=True)
    assert res == [4, 6, 8, 10, 12]


def test_tqdm_streaming():
    # Note, this doesn't work properly in PyCharm due to incomplete
    # support for carriage returns in the PyCharm console.
    cluster = get_test_cluster()
    print_fn = rh.function(fn=do_tqdm_printing_and_logging, system=cluster)
    res = print_fn(steps=40, stream_logs=True)
    assert res == list(range(50))


def test_cancel_run():
    cluster = get_test_cluster()
    print_fn = rh.function(fn=do_printing_and_logging, system=cluster)
    key = print_fn.remote()
    assert isinstance(key, str)
    res = cluster.cancel(key)
    assert res == "Cancelled"
    try:
        cluster.get(key, stream_logs=True)
    except Exception as e:
        assert "This task or its dependency was cancelled by" in str(e)


if __name__ == "__main__":
    unittest.main()



from pathlib import Path
from typing import Any, Dict, List, Optional

import ray

import yaml


def _current_cluster(key="name"):
    """Retrive key value from the current cluster config.
    If key is "config", returns entire config."""
    if Path("~/.rh/cluster_config.yaml").expanduser().exists():
        with open(Path("~/.rh/cluster_config.yaml").expanduser()) as f:
            cluster_config = yaml.safe_load(f)
        if key == "config":
            return cluster_config
        elif key == "cluster_name":
            return cluster_config["name"].rsplit("/", 1)[-1]
        return cluster_config[key]
    else:
        return None


THIS_CLUSTER = _current_cluster("cluster_name")


class ObjStore:
    """Class to handle object storage for Runhouse. Object storage for a cluster is
    stored in the Ray GCS, if available."""

    # Note, if we turn this into a ray actor we could get it by a string on server restart, which
    # would allow us to persist the store across server restarts. Unclear if this is desirable yet.
    # https://docs.ray.io/en/latest/ray-core/package-ref.html#ray-get-actor

    RH_LOGFILE_PATH = Path.home() / ".rh/logs"

    def __init__(self, cluster_name: Optional[str] = None):
        self.cluster_name = cluster_name or THIS_CLUSTER
        self._obj_store_cache = {}
        self.imported_modules = {}

    @property
    def obj_store_cache(self):
        return self._obj_store_cache

    @obj_store_cache.setter
    def obj_store_cache(self, value: Dict):
        self._obj_store_cache = value

    def put(self, key: str, value: Any):
        obj_ref = ray.put(value)
        self.obj_store_cache[key] = obj_ref

    def put_obj_ref(self, key, obj_ref):
        self.obj_store_cache[key] = obj_ref

    def get(
        self, key: str, default: Optional[Any] = None, timeout: Optional[float] = None
    ):
        obj_ref = self.obj_store_cache.get(key, None)
        if obj_ref:
            return ray.get(obj_ref, timeout=timeout)
        else:
            return default

    def get_obj_refs_list(self, keys: List):
        return [
            self.obj_store_cache.get(key, key) if isinstance(key, str) else key
            for key in keys
        ]

    def get_obj_refs_dict(self, d: Dict):
        return {
            k: self.obj_store_cache.get(v, v) if isinstance(v, str) else v
            for k, v in d.items()
        }

    def keys(self):
        return list(self.obj_store_cache.keys())

    def delete(self, key: str):
        self.obj_store_cache.pop(key, None)

    def pop(self, key: str, default: Optional[Any] = None):
        return self.obj_store_cache.pop(key, default)

    def clear(self):
        self.obj_store_cache = {}

    def cancel(self, key: str, force: bool = False, recursive: bool = True):
        obj_ref = self.obj_store_cache.get(key, None)
        if obj_ref:
            ray.cancel(obj_ref, force=force, recursive=recursive)

    def get_logfiles(self, key: str, log_type=None):
        # Info on ray logfiles: https://docs.ray.io/en/releases-2.2.0/ray-observability/ray-logging.html#id1
        obj_ref = self.obj_store_cache.get(key, None)
        if obj_ref:
            # Logs are like worker-[worker_id]-[job_id]-[pid].[out|err]
            key_logs_path = Path(self.RH_LOGFILE_PATH) / key
            # stdout_files = ray_logs_path.glob(f'worker-*-{obj_ref.job_id().hex()}-*.out')
            suffix = (
                ".out"
                if log_type == "stdout"
                else ".err"
                if log_type == "stderr"
                else ""
            )
            return [str(f.absolute()) for f in key_logs_path.glob(f"worker*{suffix}")]
        else:
            return None

    def __repr__(self):
        return f"ObjStore({self.obj_store_cache})"

    def __str__(self):
        return f"ObjStore({self.obj_store_cache})"



import logging

import pprint
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import requests

from runhouse.rh_config import rns_client
from runhouse.rns.api_utils.resource_access import ResourceAccess
from runhouse.rns.api_utils.utils import load_resp_content, read_resp_data
from runhouse.rns.top_level_rns_fns import (
    resolve_rns_path,
    save,
    split_rns_name_and_path,
)

logger = logging.getLogger(__name__)


class Resource:
    RESOURCE_TYPE = None

    def __init__(
        self,
        name: Optional[str] = None,
        dryrun: bool = None,
    ):
        """
        Runhouse abstraction for objects that can be saved, shared, and reused.

        Runhouse currently supports the following builtin Resource types:

        - Compute Abstractions
            - Cluster :py:class:`.hardware.cluster.Cluster`
            - Function :py:class:`.function.Function`
            - Package :py:class:`.packages.package.Package`


        - Data Abstractions
            - Blob :py:class:`.blob.Blob`
            - Folder :py:class:`.folders.folder.Folder`
            - Table :py:class:`.tables.table.Table`
        """
        self._name, self._rns_folder = None, None
        if name is not None:
            # TODO validate that name complies with a simple regex
            if name.startswith("/builtins/"):
                name = name[10:]
            if name[0] == "^" and name != "^":
                name = name[1:]
            self._name, self._rns_folder = rns_client.split_rns_name_and_path(
                rns_client.resolve_rns_path(name)
            )

        self.dryrun = dryrun

    # TODO add a utility to allow a parameter to be specified as "default" and then use the default value

    @property
    def config_for_rns(self):
        config = {
            "name": self.rns_address,
            "resource_type": self.RESOURCE_TYPE,
            "resource_subtype": self.__class__.__name__,
        }
        return config

    def _resource_string_for_subconfig(self, resource):
        """Returns a string representation of a sub-resource for use in a config."""
        if resource is None or isinstance(resource, str):
            return resource
        if resource.name:
            if resource.rns_address.startswith("^"):
                # Calls save internally and puts the resource in the current folder
                resource.name = rns_client.resolve_rns_path(resource.rns_address[1:])
            return resource.rns_address
        return resource.config_for_rns

    @property
    def rns_address(self):
        """Traverse up the filesystem until reaching one of the directories in rns_base_folders,
        then compute the relative path to that.

        Maybe later, account for folders along the path with a differnt RNS name."""

        if (
            self.name is None or self._rns_folder is None
        ):  # Anonymous folders have no rns address
            return None

        return str(Path(self._rns_folder) / self.name)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        # Split the name and rns path if path is given (concat with current_folder if just stem is given)
        self._name, self._rns_folder = split_rns_name_and_path(resolve_rns_path(name))

    @rns_address.setter
    def rns_address(self, new_address):
        self.name = new_address  # Note, this saves the resource to the new address!

    def _save_sub_resources(self):
        """Overload by child resources to save any resources they hold internally."""
        pass

    def save(
        self,
        name: str = None,
        overwrite: bool = True,
    ):
        """Register the resource, saving it to local working_dir config and RNS config store. Uses the resource's
        `self.config_for_rns` to generate the dict to save."""

        self._save_sub_resources()
        # TODO deal with logic of saving anonymous folder for the first time after naming, i.e.
        # Path(tempfile.gettempdir()).relative_to(self.path) ...
        if name:
            if "/" in name[1:] or self._rns_folder is None:
                self._name, self._rns_folder = split_rns_name_and_path(
                    resolve_rns_path(name)
                )
            else:
                self._name = name

        # TODO handle self.access == 'read' instead of this weird overwrite argument
        save(self, overwrite=overwrite)

        return self

    def __str__(self):
        return pprint.pformat(self.config_for_rns)

    @classmethod
    def _check_for_child_configs(cls, config):
        """Overload by child resources to load any resources they hold internally."""
        return config

    @classmethod
    def from_name(cls, name, dryrun=True):
        """Load existing Resource via its name."""
        config = rns_client.load_config(name=name)
        if not config:
            raise ValueError(f"Resource {name} not found.")
        config["name"] = name
        config = cls._check_for_child_configs(config)
        # Uses child class's from_config
        return cls.from_config(config=config, dryrun=dryrun)

    def unname(self):
        """Remove the name of the resource. This changes the resource name to anonymous and deletes any local
        or RNS configs for the resource."""
        self.delete_configs()
        self._name = None

    @staticmethod
    def history(name: str) -> List[Dict]:
        """Return the history of the resource, including specific config fields (e.g. blob path) and which runs
        have overwritten it."""
        resource_uri = rns_client.resource_uri(name)
        resp = requests.get(
            f"{rns_client.api_server_url}/resource/history/{resource_uri}",
            headers=rns_client.request_headers,
        )
        if resp.status_code != 200:
            raise Exception(
                f"Failed to load resource history: {load_resp_content(resp)}"
            )

        resource_history = read_resp_data(resp)
        return resource_history

    # TODO delete sub-resources
    def delete_configs(self):
        """Delete the resource's config from local working_dir and RNS config store."""
        rns_client.delete_configs(resource=self)

    def save_attrs_to_config(self, config: Dict, attrs: List[str]):
        """Save the given attributes to the config"""
        for attr in attrs:
            val = self.__getattribute__(attr)
            if val:
                config[attr] = val

    def is_local(self):
        return (
            hasattr(self, "install_target")
            and isinstance(self.install_target, str)
            and self.install_target.startswith("~")
            or hasattr(self, "system")
            and self.system == "file"
        )

    # TODO [DG] Implement proper sharing of subresources (with an overload of some kind)
    def share(
        self,
        users: list,
        access_type: Union[ResourceAccess, str] = ResourceAccess.READ,
        notify_users: bool = False,
    ) -> Tuple[Dict[str, ResourceAccess], Dict[str, ResourceAccess]]:
        """Grant access to the resource for the list of users. If a user has a Runhouse account they
        will receive an email notifying them of their new access. If the user does not have a Runhouse account they will
        also receive instructions on creating one, after which they will be able to have access to the Resource.

        .. note::
            You can only grant resource access to other users if you have Write / Read privileges for the Resource.

        Args:
            users (list): list of user emails and / or runhouse account usernames.
            access_type (:obj:`ResourceAccess`, optional): access type to provide for the resource.
            notify_users (bool): Send email notification to users who have been given access. Defaults to `False`.

        Returns:
            `added_users`: users who already have an account and have been granted access to the resource.
            `new_users`: users who do not have Runhouse accounts.

        Example:
            >>> added_users, new_users = my_resource.share(users=["username1", "user2@gmail.com"], access_type='write')
        """
        if self.name is None:
            raise ValueError("Resource must have a name in order to share")

        if hasattr(self, "system") and self.system in ["ssh", "sftp"]:
            logger.warning(
                "Sharing a resource located on a cluster is not recommended. For persistence, we suggest"
                "saving to a cloud storage system (ex: `s3` or `gs`). You can copy your cluster based "
                f"{self.RESOURCE_TYPE} to your desired storage provider using the `.to()` method. "
                f"For example: `{self.RESOURCE_TYPE}.to(system='rh-cpu')`"
            )

        if self.is_local():
            if self.RESOURCE_TYPE == "package":
                raise TypeError(
                    f"Unable to share a local {self.RESOURCE_TYPE}. Please make sure the {self.RESOURCE_TYPE} is "
                    f"located on a cluster. You can use the `.to_cluster()` method to do so. "
                    f"For example: `{self.name}.to_cluster(system='rh-cpu')`"
                )
            else:
                raise TypeError(
                    f"Unable to share a local {self.RESOURCE_TYPE}. Please make sure the {self.RESOURCE_TYPE} is "
                    f"located on a cluster or a remote system. You can use the `.to()` method to do so. "
                    f"For example: `{self.name}.to(system='s3')`"
                )

        if isinstance(access_type, str):
            access_type = ResourceAccess(access_type)

        if not rns_client.exists(self.rns_address):
            self.save(name=rns_client.local_to_remote_address(self.rns_address))

        added_users, new_users = rns_client.grant_resource_access(
            rns_address=self.rns_address,
            user_emails=users,
            access_type=access_type,
            notify_users=notify_users,
        )
        return added_users, new_users



import logging

from .. import OnDemandCluster

from .table import Table

logger = logging.getLogger(__name__)


class DaskTable(Table):
    DEFAULT_FOLDER_PATH = "/runhouse/dask-tables"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def from_config(config: dict, dryrun=True):
        """Load config values into the object."""
        if isinstance(config["system"], dict):
            config["system"] = OnDemandCluster.from_config(
                config["system"], dryrun=dryrun
            )
        return DaskTable(**config)

    def write(self, write_index: bool = False):
        # https://docs.dask.org/en/stable/how-to/connect-to-remote-data.html
        if self._cached_data is not None:
            # https://stackoverflow.com/questions/72891631/how-to-remove-null-dask-index-from-parquet-file
            self.data.to_parquet(
                self.fsspec_url,
                write_index=write_index,
                storage_options=self.data_config,
            )
            logger.info(f"Saved {str(self)} to: {self.fsspec_url}")

        return self

    def fetch(self, **kwargs):
        import dask.dataframe as dd

        # https://docs.dask.org/en/stable/generated/dask.dataframe.read_parquet.html
        self._cached_data = dd.read_parquet(
            self.fsspec_url, storage_options=self.data_config
        )

        return self._cached_data



import contextlib
import logging
import pkgutil
import subprocess
import threading
import time
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import grpc
import ray.cloudpickle as pickle
import sshtunnel

from sky.utils import command_runner
from sshtunnel import HandlerSSHTunnelForwarderError, SSHTunnelForwarder

from runhouse.rh_config import open_grpc_tunnels, rns_client
from runhouse.rns.packages.package import Package

from runhouse.rns.resource import Resource

from runhouse.servers.grpc.unary_client import UnaryClient
from runhouse.servers.grpc.unary_server import UnaryService

logger = logging.getLogger(__name__)


class Cluster(Resource):
    RESOURCE_TYPE = "cluster"
    GRPC_TIMEOUT = 5  # seconds

    def __init__(
        self,
        name,
        ips: List[str] = None,
        ssh_creds: Dict = None,
        dryrun=True,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        The Runhouse cluster, or system. This is where you can run Functions or access/transfer data
        between. You can BYO (bring-your-own) cluster by providing cluster IP and ssh_creds, or
        this can be an on-demand cluster that is spun up/down through
        `SkyPilot <https://github.com/skypilot-org/skypilot>`_, using your cloud credentials.

        .. note::
            To build a cluster, please use the factory method :func:`cluster`.
        """

        super().__init__(name=name, dryrun=dryrun)

        self.address = ips[0] if ips else None
        self._ssh_creds = ssh_creds
        self.ips = ips
        self._grpc_tunnel = None
        self.client = None

        if not dryrun and self.address:
            # OnDemandCluster will start ray itself, but will also set address later, so won't reach here.
            self.start_ray()
            self.save_config_to_cluster()

    def save_config_to_cluster(self):
        import json

        config = self.config_for_rns
        if "sky_state" in config.keys():
            # a bunch of setup commands that mess up json dump
            del config["sky_state"]
        json_config = f"{json.dumps(config)}"

        self.run(
            [
                f"mkdir -p ~/.rh; touch ~/.rh/cluster_config.yaml; echo '{json_config}' > ~/.rh/cluster_config.yaml"
            ]
        )

    @staticmethod
    def from_config(config: dict, dryrun=False):
        # TODO use 'resource_subtype' in config?
        if "ips" in config:
            return Cluster(**config, dryrun=dryrun)
        else:
            from runhouse.rns.hardware import OnDemandCluster

            return OnDemandCluster(**config, dryrun=dryrun)

    @property
    def config_for_rns(self):
        config = super().config_for_rns
        self.save_attrs_to_config(config, ["ips"])
        if self.ips is not None:
            config["ssh_creds"] = self.ssh_creds()
        return config

    def is_up(self) -> bool:
        """Check if the cluster is up."""
        return self.address is not None

    def up_if_not(self):
        """Bring up the cluster if it is not up. No-op if cluster is already up."""
        if not self.is_up():
            if not hasattr(self, "up"):
                raise NotImplementedError(
                    f"Cluster <{self.name}> does not have an up method."
                )
            self.up()
        return self

    def keep_warm(self):
        logger.info(
            f"cluster.keep_warm will have no effect on self-managed cluster {self.name}."
        )

    def start_ray(self):
        if self.is_up():
            res = self.run(["ray start --head"], stream_logs=False)
            if res[0] == 0:
                return
            if any(
                line.startswith("ConnectionError: Ray is trying to start at")
                for line in res[0][1].splitlines()
            ):
                # Ray is already started
                return
            # Check if ray is installed
            if "ray" not in self._get_pip_installs(strip_versions=True):
                self.run(["pip install ray"])
                res = self.run(["ray start --head"])
                if not res[0][0]:
                    raise RuntimeError(
                        f"Failed to start ray on cluster <{self.name}>. "
                        f"Error: {res[0][1]}"
                    )
        else:
            raise ValueError(f"Cluster <{self.name}> is not up.")

    def _get_pip_installs(self, strip_versions=False):
        packages = self.run(["pip freeze"], stream_logs=False)[0][1].splitlines()
        if strip_versions:
            packages = [p.split("==")[0] for p in packages]
        return packages

    def sync_runhouse_to_cluster(self, _install_url=None):
        if not self.address:
            raise ValueError(f"No address set for cluster <{self.name}>. Is it up?")
        local_rh_package_path = Path(pkgutil.get_loader("runhouse").path).parent

        # Check if runhouse is installed from source and has setup.py
        if (
            not _install_url
            and local_rh_package_path.parent.name == "runhouse"
            and (local_rh_package_path.parent / "setup.py").exists()
        ):
            # Package is installed in editable mode
            local_rh_package_path = local_rh_package_path.parent
            rh_package = Package.from_string(
                f"reqs:{local_rh_package_path}", dryrun=True
            )
            rh_package.to_cluster(self, mount=False)
            status_codes = self.run(["pip install ./runhouse"], stream_logs=True)
        # elif local_rh_package_path.parent.name == 'site-packages':
        else:
            # Package is installed in site-packages
            # status_codes = self.run(['pip install runhouse-nightly==0.0.2.20221202'], stream_logs=True)
            # rh_package = 'runhouse_nightly-0.0.1.dev20221202-py3-none-any.whl'
            # rh_download_cmd = f'curl https://runhouse-package.s3.amazonaws.com/{rh_package} --output {rh_package}'
            # TODO need to check user's current version and install same version?
            _install_url = _install_url or "runhouse"
            rh_install_cmd = f"pip install {_install_url}"
            status_codes = self.run([rh_install_cmd], stream_logs=True)

        if status_codes[0][0] != 0:
            raise ValueError(f"Error installing runhouse on cluster <{self.name}>")

    def install_packages(self, reqs: List[Union[Package, str]]):
        """Install the given packages on the cluster."""
        self.check_grpc()
        to_install = []
        for package in reqs:
            if isinstance(package, str):
                # If the package is a local folder, we need to create the package to sync it over to the cluster
                pkg_obj = Package.from_string(package, dryrun=False)
            else:
                pkg_obj = package

            from runhouse.rns.folders.folder import Folder

            if (
                isinstance(pkg_obj.install_target, Folder)
                and pkg_obj.install_target.is_local()
            ):
                pkg_str = pkg_obj.name or Path(pkg_obj.install_target.path).name
                logging.info(
                    f"Copying local package {pkg_str} to cluster <{self.name}>"
                )
                remote_package = pkg_obj.to_cluster(self, mount=False)
                to_install.append(remote_package)
            else:
                to_install.append(package)  # Just appending the string!
        logging.info(
            f"Installing packages on cluster {self.name}: "
            f"{[req if isinstance(req, str) else str(req) for req in reqs]}"
        )
        self.client.install_packages(to_install)

    def get(self, key: str, default: Any = None, stream_logs: bool = False):
        """Get the object at the given key from the cluster's object store."""
        self.check_grpc()
        return self.client.get_object(key, stream_logs=stream_logs) or default

    def add_secrets(self, provider_secrets: dict):
        """Copy secrets from current environment onto the cluster"""
        self.check_grpc()
        return self.client.add_secrets(pickle.dumps(provider_secrets))

    def put(self, key: str, obj: Any):
        """Put the given object on the cluster's object store at the given key."""
        self.check_grpc()
        return self.client.put_object(key, obj)

    # TODO [DG] add a method to list all the keys in the cluster

    def cancel(self, key, force=False):
        """Cancel the given run on cluster."""
        self.check_grpc()
        return self.client.cancel_runs(key, force=force)

    def clear_pins(self, pins: Optional[List[str]] = None):
        """Remove the given pinned items from the cluster. If `pins` is set to ``None``, then
        all pinned objects will be cleared."""
        self.check_grpc()
        self.client.clear_pins(pins)
        logger.info(f'Clearing pins on cluster {pins or ""}')

    def on_same_cluster(self, resource: Resource):
        """Whether the given resource is on the cluster."""
        if hasattr(resource, "system") and isinstance(resource.system, Resource):
            return resource.system.rns_address == self.rns_address
        return False

    # ----------------- gRPC Methods ----------------- #

    def connect_grpc(self, force_reconnect=False):
        # FYI based on: https://sshtunnel.readthedocs.io/en/latest/#example-1
        # FYI If we ever need to do this from scratch, we can use this example:
        # https://github.com/paramiko/paramiko/blob/main/demos/rforward.py#L74
        if not self.address:
            raise ValueError(f"No address set for cluster <{self.name}>. Is it up?")

        # TODO [DG] figure out how to ping to see if tunnel is already up
        if self._grpc_tunnel and force_reconnect:
            self._grpc_tunnel.close()

        # TODO Check if port is already open instead of refcounting?
        # status = subprocess.run(['nc', '-z', self.address, str(self.grpc_port)], capture_output=True)
        # if not self.check_port(self.address, UnaryClient.DEFAULT_PORT):

        tunnel_refcount = 0
        if self.address in open_grpc_tunnels:
            ssh_tunnel, connected_port, tunnel_refcount = open_grpc_tunnels[
                self.address
            ]
            ssh_tunnel.check_tunnels()
            if ssh_tunnel.tunnel_is_up[ssh_tunnel.local_bind_address]:
                self._grpc_tunnel = ssh_tunnel
        else:
            self._grpc_tunnel, connected_port = self.ssh_tunnel(
                UnaryClient.DEFAULT_PORT,
                remote_port=UnaryService.DEFAULT_PORT,
                num_ports_to_try=5,
            )
        open_grpc_tunnels[self.address] = (
            self._grpc_tunnel,
            connected_port,
            tunnel_refcount + 1,
        )

        # Connecting to localhost because it's tunneled into the server at the specified port.
        self.client = UnaryClient(host="127.0.0.1", port=connected_port)
        waited = 0
        while not self.is_connected() and waited <= self.GRPC_TIMEOUT:
            time.sleep(0.25)
            waited += 0.25

    def check_grpc(self, restart_grpc_server=True):
        if not self.address:
            # For OnDemandCluster, this initial check doesn't trigger a sky.status, which is slow.
            # If cluster simply doesn't have an address we likely need to up it.
            if not hasattr(self, "up"):
                raise ValueError(
                    "Cluster must have an ip address (i.e. be up) or have a reup_cluster method "
                    "(e.g. OnDemandCluster)."
                )
            if not self.is_up():
                # If this is a OnDemandCluster, before we up the cluster, run a sky.status to see if the cluster
                # is already up but doesn't have an address assigned yet.
                self.up_if_not()

        if not self.client:
            try:
                self.connect_grpc()
            except (
                grpc.RpcError,
                sshtunnel.BaseSSHTunnelForwarderError,
            ):
                # It's possible that the cluster went down while we were trying to install packages.
                if not self.is_up():
                    self.up_if_not()
                else:
                    self.restart_grpc_server(resync_rh=False)

        if self.is_connected():
            return

        self.connect_grpc()
        if self.is_connected():
            return

        if restart_grpc_server:
            self.restart_grpc_server(resync_rh=False)
            self.connect_grpc()
            if self.is_connected():
                return

            self.restart_grpc_server(resync_rh=True)
            self.connect_grpc()
            if self.is_connected():
                return

        raise ValueError(f"Could not connect to cluster <{self.name}>")

        # try:
        #     self.client.ping()
        # except Exception as e:
        #     if restart_if_down:
        #         self.restart_grpc_server(resync_rh=resync_rh)
        #         self.connect_grpc(force_reconnect=True)
        #         self.client.ping()
        #     else:
        #         raise e

    @staticmethod
    def check_port(ip_address, port):
        import socket

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        return s.connect_ex(("127.0.0.1", int(port)))

    def ssh_tunnel(
        self, local_port, remote_port=None, num_ports_to_try: int = 0
    ) -> Tuple[SSHTunnelForwarder, int]:
        # Debugging cmds (mac):
        # netstat -vanp tcp | grep 5005
        # lsof -i :5005_
        # kill -9 <pid>

        creds: dict = self.ssh_creds()
        connected = False
        ssh_tunnel = None
        while not connected:
            try:
                if local_port > local_port + num_ports_to_try:
                    raise Exception(
                        f"Failed to create SSH tunnel after {num_ports_to_try} attempts"
                    )

                ssh_tunnel = SSHTunnelForwarder(
                    self.address,
                    ssh_username=creds["ssh_user"],
                    ssh_pkey=creds["ssh_private_key"],
                    local_bind_address=("", local_port),
                    remote_bind_address=("127.0.0.1", remote_port or local_port),
                    set_keepalive=1,
                    # mute_exceptions=True,
                )
                ssh_tunnel.start()
                connected = True
            except HandlerSSHTunnelForwarderError:
                # try connecting with a different port - most likely the issue is the port is already taken
                local_port += 1
                num_ports_to_try -= 1
                pass

        return ssh_tunnel, local_port

    # TODO [DG] Remove this for now, for some reason it was causing execution to hang after programs completed
    # def __del__(self):
    #     if self.address in open_grpc_tunnels:
    #         tunnel, port, refcount = open_grpc_tunnels[self.address]
    #         if refcount == 1:
    #             tunnel.close()
    #             open_grpc_tunnels.pop(self.address)
    #         else:
    #             open_grpc_tunnels[self.address] = (tunnel, port, refcount - 1)
    #     elif self._grpc_tunnel:  # Not sure why this would be reached but keeping it just in case
    #         self._grpc_tunnel.close()

    # import paramiko
    # ssh = paramiko.SSHClient()
    # ssh.load_system_host_keys()
    # ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    # from pathlib import Path
    # ssh.connect(self.address,
    #             username=creds['ssh_user'],
    #             key_filename=str(Path(creds['ssh_private_key']).expanduser())
    #             )
    # transport = ssh.get_transport()
    # transport.request_port_forward('', local_port)
    # ssh_tunnel = transport.open_channel("direct-tcpip", ("localhost", local_port),
    #                                     (self.address, remote_port or local_port))
    # if ssh_tunnel.is_active():
    #     connected = True
    #     print(f"SSH tunnel is open to {self.address}:{local_port}")

    def restart_grpc_server(
        self,
        _rh_install_url: str = None,
        resync_rh: bool = True,
        restart_ray: bool = False,
    ):
        """Restart the GRPC server."""
        # TODO how do we capture errors if this fails?
        if resync_rh:
            self.sync_runhouse_to_cluster(_install_url=_rh_install_url)
            self.save_config_to_cluster()
        kill_proc_cmd = 'pkill -f "python3 -m runhouse.servers.grpc.unary_server"'
        logfile = f"{self.name}_grpc_server.log"
        grpc_server_cmd = "python3 -m runhouse.servers.grpc.unary_server"
        # 2>&1 redirects stderr to stdout
        screen_cmd = f"screen -dm bash -c '{grpc_server_cmd} >> ~/.rh/{logfile} 2>&1'"
        cmds = [kill_proc_cmd]
        if restart_ray:
            cmds.append("ray stop")
            cmds.append(
                "ray start --head"
            )  # Need to set gpus or Ray will block on cpu-only clusters
        cmds.append(screen_cmd)

        # If we need different commands for debian or ubuntu, we can use this:
        # Need to get actual provider in case provider == 'cheapest'
        # handle = sky.global_user_state.get_cluster_from_name(self.name)['handle']
        # cloud_provider = str(handle.launched_resources.cloud)
        # ubuntu_kill_proc_cmd = f'fuser -k {UnaryService.DEFAULT_PORT}/tcp'
        # debian_kill_proc_cmd = "kill -9 $(netstat -anp | grep 50052 | grep -o '[0-9]*/' | sed 's+/$++')"
        # f'kill -9 $(lsof -t -i:{UnaryService.DEFAULT_PORT})'
        # kill_proc_at_port_cmd = debian_kill_proc_cmd if cloud_provider == 'GCP' \
        #     else ubuntu_kill_proc_cmd

        status_codes = self.run(
            commands=cmds,
            stream_logs=True,
        )
        # As of 2022-27-Dec still seems we need this.
        time.sleep(2)
        return status_codes

    @contextlib.contextmanager
    def pause_autostop(self):
        """Context manager to temporarily pause autostop. Mainly for OnDemand clusters, for BYO cluster
        there is no autostop."""
        pass

    def run_module(self, relative_path, module_name, fn_name, fn_type, args, kwargs):
        self.check_grpc()
        return self.client.run_module(
            relative_path, module_name, fn_name, fn_type, args, kwargs
        )

    def is_connected(self):
        return self.client is not None and self.client.is_connected()

    def disconnect(self):
        if self._grpc_tunnel:
            self._grpc_tunnel.stop()
        # if self.client:
        #     self.client.shutdown()

    def __getstate__(self):
        """Delete non-serializable elements (e.g. thread locks) before pickling."""
        state = self.__dict__.copy()
        state["client"] = None
        state["_grpc_tunnel"] = None
        return state

    # ----------------- SSH Methods ----------------- #

    def ssh_creds(self):
        return self._ssh_creds

    def rsync(self, source: str, dest: str, up: bool, contents: bool = False):
        """
        Sync the contents of the source directory into the destination.

        .. note:
            Ending `source` with a slash will copy the contents of the directory into dest,
            while omitting it will copy the directory itself (adding a directory layer)."""
        # FYI, could be useful: https://github.com/gchamon/sysrsync
        if contents:
            source = source + "/" if not source.endswith("/") else source
            dest = dest + "/" if not dest.endswith("/") else dest
        ssh_credentials = self.ssh_creds()
        runner = command_runner.SSHCommandRunner(self.address, **ssh_credentials)
        if up:
            runner.run(["mkdir", "-p", dest], stream_logs=False)
        else:
            Path(dest).expanduser().parent.mkdir(parents=True, exist_ok=True)
        runner.rsync(source, dest, up=up, stream_logs=False)

    def ssh(self):
        creds = self.ssh_creds()
        subprocess.run(
            f"ssh {creds['ssh_user']}:{self.address} -i {creds['ssh_private_key']}".split(
                " "
            )
        )

    def ping(self, timeout=5):
        ssh_call = threading.Thread(target=lambda: self.run(['echo "hello"']))
        ssh_call.start()
        ssh_call.join(timeout=timeout)
        if ssh_call.is_alive():
            raise TimeoutError("SSH call timed out")
        return True

    def run(
        self,
        commands: List[str],
        stream_logs: bool = True,
        port_forward: Optional[int] = None,
        require_outputs: bool = True,
    ):
        """Run a list of shell commands on the cluster."""
        # TODO [DG] suspect autostop while running?
        runner = command_runner.SSHCommandRunner(self.address, **self.ssh_creds())
        return_codes = []
        for command in commands:
            logger.info(f"Running command on {self.name}: {command}")
            ret_code = runner.run(
                command,
                require_outputs=require_outputs,
                stream_logs=stream_logs,
                port_forward=port_forward,
            )
            return_codes.append(ret_code)
        return return_codes

    def run_python(
        self,
        commands: List[str],
        stream_logs: bool = True,
        port_forward: Optional[int] = None,
    ):
        """Run a list of python commands on the cluster."""
        command_str = "; ".join(commands)
        return_codes = self.run(
            [f'python3 -c "{command_str}"'],
            stream_logs=stream_logs,
            port_forward=port_forward,
        )
        return return_codes

    def send_secrets(self, providers: Optional[List[str]] = None):
        """Send secrets for the given providers. If none provided will send secrets for providers that have been
        configured in the environment."""
        from runhouse import Secrets

        Secrets.to(hardware=self, providers=providers)

    def ipython(self):
        # TODO tunnel into python interpreter in cluster
        pass

    def notebook(
        self,
        persist: bool = False,
        sync_package_on_close: Optional[str] = None,
        port_forward: int = 8888,
    ):
        """Tunnel into and launch notebook from the cluster."""
        tunnel, port_fwd = self.ssh_tunnel(local_port=port_forward, num_ports_to_try=10)
        try:
            install_cmd = "pip install jupyterlab"
            jupyter_cmd = f"jupyter lab --port {port_fwd} --no-browser"
            with self.pause_autostop():
                self.run(commands=[install_cmd, jupyter_cmd], stream_logs=True)

        finally:
            if sync_package_on_close:
                if sync_package_on_close == "./":
                    sync_package_on_close = rns_client.locate_working_dir()
                pkg = Package.from_string("local:" + sync_package_on_close)
                self.rsync(source=f"~/{pkg.name}", dest=pkg.local_path, up=False)
            if not persist:
                tunnel.stop()
                kill_jupyter_cmd = f"jupyter notebook stop {port_fwd}"
                self.run(commands=[kill_jupyter_cmd])



import logging
from typing import Optional

from .. import OnDemandCluster

from .table import Table

logger = logging.getLogger(__name__)


class HuggingFaceTable(Table):
    DEFAULT_FOLDER_PATH = "/runhouse/huggingface-tables"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def from_config(config: dict, dryrun=True):
        """Load config values into the object."""
        if isinstance(config["system"], dict):
            config["system"] = OnDemandCluster.from_config(
                config["system"], dryrun=dryrun
            )
        return HuggingFaceTable(**config, dryrun=dryrun)

    def write(self):
        hf_dataset = None
        if self._cached_data is not None:
            import datasets

            if isinstance(self.data, datasets.Dataset):
                arrow_table = self.data.data.table
                ray_dataset = self.ray_dataset_from_arrow(arrow_table)
                self.data, hf_dataset = ray_dataset, self.data
                self.write_ray_dataset(self.data)
            elif isinstance(self.data, datasets.DatasetDict):
                # TODO [JL] Add support for dataset dict
                raise NotImplementedError(
                    "Runhouse does not currently support DatasetDict objects, please convert to "
                    "a Dataset before saving."
                )
            else:
                raise TypeError(
                    "Unsupported data type for HuggingFaceTable. Please use a Dataset"
                )

            logger.info(f"Saved {str(self)} to: {self.fsspec_url}")

        # Restore the original dataset
        if hf_dataset is not None:
            self.data = self.to_dataset(hf_dataset)

        return self

    def fetch(self, **kwargs):
        # Read as pyarrow table, then convert back to HF dataset
        arrow_table = super().fetch(**kwargs)
        self._cached_data = self.to_dataset(arrow_table)
        return self._cached_data

    def stream(
        self,
        batch_size: int,
        drop_last: bool = False,
        shuffle_seed: Optional[int] = None,
        shuffle_buffer_size: Optional[int] = None,
        prefetch_blocks: Optional[int] = None,
        as_dict: bool = True,
    ):
        """Stream data as either Dataset object or dict (as generated by ray iter)"""
        for batch in super().stream(
            batch_size, drop_last, shuffle_seed, shuffle_buffer_size, prefetch_blocks
        ):
            yield batch if as_dict else self.to_dataset(batch)

    @staticmethod
    def to_dataset(data):
        """Convert to a HuggingFace Dataset. Relevant when fetching the data or when choosing to stream the data
        in as a Dataset."""
        import pandas as pd
        import pyarrow as pa
        import ray.data
        from datasets import Dataset

        if isinstance(data, Dataset):
            return data

        elif isinstance(data, dict):
            return Dataset.from_dict(data)

        elif isinstance(data, pd.DataFrame):
            return Dataset.from_pandas(data)

        elif isinstance(data, (pa.Table, ray.data.Dataset)):
            return Dataset.from_pandas(data.to_pandas())

        else:
            raise TypeError(
                f"Data must be a dict, Pandas DataFrame, ray Dataset, or PyArrow table, not {type(data)}"
            )
